;;; $Id: lambda-if-exp.scm,v 1.4 2006/01/05 22:24:09 leavens Exp $
;;; An ADT for the untyped multiple-argument lambda calculus,
;;; with quoted symbols and if expressions.
;;;
;;; Grammar:
;;;  <expression> ::=
;;;       <identifier>                                "var-exp (id)"
;;;     | (quote <identifier>)                        "quote-exp (symbol)"
;;;     | ( lambda ( {<identifier>}* ) <expression> ) "lambda-exp (ids body)"
;;;     | ( if <expression>                           "if-exp (test-exp
;;;            <expression>                                    true-exp
;;;            <expression> )                                  false-exp)"
;;;     | ( <expression> {<expression>}* )            "app-exp (rator rands)"

(module lambda-if-exp (lib "typedscm.ss" "typedscm")

(provide expression?
         var-exp? quote-exp? lambda-exp? if-exp? app-exp?
         var-exp->id  quote-exp->symbol lambda-exp->ids
         lambda-exp->body app-exp->rator app-exp->rands
         if-exp->test-exp if-exp->true-exp if-exp->false-exp
         var-exp quote-exp lambda-exp if-exp
         app-exp parse-expression)

;; type predicate
(deftype expression? (type-predicate-for expression))

;; case testers (discriminators)
(deftype var-exp? (-> (expression) boolean))
(deftype quote-exp? (-> (expression) boolean))
(deftype lambda-exp? (-> (expression) boolean))
(deftype if-exp? (-> (expression) boolean))
(deftype app-exp? (-> (expression) boolean))

;; observers 
(deftype var-exp->id (-> (expression) symbol))
(deftype quote-exp->symbol (-> (expression) symbol))
(deftype lambda-exp->ids (-> (expression) (list-of symbol)))
(deftype lambda-exp->body (-> (expression) expression))
(deftype app-exp->rator (-> (expression) expression))
(deftype app-exp->rands (-> (expression) (list-of expression)))
(deftype if-exp->test-exp (-> (expression) expression))
(deftype if-exp->true-exp (-> (expression) expression))
(deftype if-exp->false-exp (-> (expression) expression))

;; constructors
(deftype var-exp (-> (symbol) expression))
(deftype quote-exp (-> (symbol) expression))
(deftype lambda-exp (-> ((list-of symbol) expression) expression))
(deftype if-exp (-> (expression expression expression) expression))
(deftype app-exp (-> (expression (list-of expression)) expression))

;; parsing/bless input correctness
(deftype parse-expression (-> (datum) expression))

(require (lib "all-mod.scm" "lib342"))

(defrep (expression datum))

(define expression?
  (lambda (exp)
    (has-type-trusted
      boolean
      (or (var-exp? exp)
          (and (quote-exp? exp) 
               (symbol? (quote-exp->symbol exp)))
          (and (lambda-exp? exp) 
               (expression? (lambda-exp->body exp)))
          (and (if-exp? exp)
               (expression? (if-exp->test-exp exp))
               (expression? (if-exp->true-exp exp))
               (expression? (if-exp->false-exp exp)))
          (and (app-exp? exp)
               (expression? (app-exp->rator exp))
               (all expression? (app-exp->rands exp)))))))

(define var-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents a var-exp
    (symbol? exp)))

(define quote-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents a quote-exp
    (test-type
      boolean?
      (and (list? exp)
           (= (length exp) 2)
           (eq? (car exp) 'quote)
           (symbol? (cadr exp))))))

(define lambda-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents a lambda-exp
    (test-type
      boolean?
      (and (list? exp)
           (= (length exp) 3)
           (eq? (car exp) 'lambda)
           (let ((ids (cadr exp)))
             (and (list? ids)
                  (all symbol? ids)))))))

(define if-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents an if-exp
    (test-type
      boolean?
      (and (list? exp)
           (= (length exp) 4)
           (eq? (car exp) 'if)))))

(define app-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents an app-exp
    (test-type
      boolean?
      (and (list? exp)
           (>= (length exp) 1)
           (let ((operator (car exp)))
             (and (not (eq? operator 'quote))
                  (not (eq? operator 'lambda))
                  (not (eq? operator 'if))))))))

(define var-exp->id
  (lambda (vr)
    ;; REQUIRES: vr represents a var-exp
    ;; ENSURES: result is the symbol represented by vr
    (if (var-exp? vr)
        (test-type symbol? vr)
        (error "var-exp->id: not a var-exp:" vr))))

(define quote-exp->symbol
  (lambda (qe)
    ;; REQUIRES: qe represents a quote-exp
    ;; ENSURES: result is the operand in the representation of qe
    (if (quote-exp? qe)
        (test-type symbol? (cadr qe))
        (error "quote-exp->symbol: not a quote-exp:" qe))))

(define lambda-exp->ids
  (lambda (lm)
    ;; REQUIRES: lm represents a lambda-exp
    ;; ENSURES: result is the list of formals in the representation of lm
    (if (lambda-exp? lm)
        (test-type (list-of symbol?) (cadr lm))
        (error "lambda-exp->id: not a lambda-exp:" lm))))

(define lambda-exp->body
  (lambda (lm)
    ;; REQUIRES: lm represents a lambda-exp
    ;; ENSURES: result is the body in the representation of lm
    (if (lambda-exp? lm)
        (test-type expression? (caddr lm))
        (error "lambda-exp->body: not a lambda-exp:" lm))))

(define app-exp->rator
  (lambda (ap)
    ;; REQUIRES: ap represents an app-exp
    ;; ENSURES: result is the operator in the representation of ap
    (if (app-exp? ap)
        (test-type expression? (car ap))
        (error "app-exp->rator: not an app-exp:" ap))))

(define app-exp->rands
  (lambda (ap)
    ;; REQUIRES: ap represents an app-exp
    ;; ENSURES: result is the list of operands in the representation of ap
    (if (app-exp? ap)
        (test-type (list-of expression?) (cdr ap))
        (error "app-exp->rands: not an app-exp:" ap))))

(define if-exp->test-exp
  (lambda (ie)
    ;; REQUIRES: ie represents an if-exp
    ;; ENSURES: result is the test expression in ie
    (if (if-exp? ie)
        (test-type expression? (cadr ie))
        (error "if-exp->test-exp: not an if-exp:" ie))))


(define if-exp->true-exp
  (lambda (ie)
    ;; REQUIRES: ie represents an if-exp
    ;; ENSURES: result is the true expression in ie
    (if (if-exp? ie)
        (test-type expression? (caddr ie))
        (error "if-exp->true-exp: not an if-exp:" ie))))

(define if-exp->false-exp
  (lambda (ie)
    ;; REQUIRES: ie represents an if-exp
    ;; ENSURES: result is the false expression in ie
    (if (if-exp? ie)
        (test-type expression? (cadddr ie))
        (error "if-exp->false-exp: not an if-exp:" ie))))

(define var-exp
  (lambda (id)
    ;; ENSURES: result is a representation of a var-exp with symbol id
    (has-type datum id)))

(define quote-exp
  (lambda (sym)
    ;; ENSURES: result is a representation of a quote-exp with symbol sym
    (has-type datum (list 'quote sym))))

(define lambda-exp
  (lambda (ids body)
    ;; ENSURES: result is a representation of a lambda-exp with formals ids
    ;;          and body body
    (has-type datum (list 'lambda ids body))))

(define if-exp
  (lambda (test-exp true-exp false-exp)
    ;; ENSURES: result is a representation of an if-exp with test
    ;; test, true expression, true-exp, and false expression false-exp
    (has-type datum (list 'if test-exp true-exp false-exp))))

(define app-exp
  (lambda (rator rands)
    ;; ENSURES: result is a representation of an app-exp with operator rator
    ;;          and operands rands
    (has-type datum (cons rator rands))))
  
(define parse-expression
  (lambda (exp)
    (test-type
      datum?
      (cond
       ((symbol? exp) (var-exp exp))
       ((quote-exp? exp) (quote-exp (cadr exp)))
       ((lambda-exp? exp)
        (lambda-exp (cadr exp) (parse-expression (caddr exp))))
       ((if-exp? exp) (if-exp (parse-expression (cadr exp))
                              (parse-expression (caddr exp))
                              (parse-expression (cadddr exp))))
       ((app-exp? exp) (app-exp (parse-expression (car exp))
                                (map parse-expression (cdr exp))))
       (else (error "parse-expression: bad syntax in expression: " exp))))))

) ;; end module
