;;; $Id: lexical-addr-exp-mod.scm,v 1.2 2006/01/05 22:24:09 leavens Exp $
;;; An ADT for lexical address expressions; this is essentially
;;; the untyped multiple-argument lambda calculus
;;; with quoted symbols and if expressions, but with lexical addresses
;;; replacing var-exps.

;;;  <lexical-addr-exp> ::=
;;;       <lexical-addr>                           "la:var-exp (lexical-addr)"
;;;     | (quote <identifier>)                     "la:quote-exp (symbol)"
;;;     | ( lambda ( {<identifier>}* )             "la:lambda-exp (ids
;;;           <lexical-addr-exp> )                                 body)"
;;;     | ( if <lexical-addr-exp>                  "la:if-exp (test-exp
;;;            <lexical-addr-exp>                              true-exp
;;;            <lexical-addr-exp> )                            false-exp)"
;;;     | ( <lexical-addr-exp>                     "la:app-exp (rator
;;;         {<lexical-addr-exp>}* )                             rands)"

(module lexical-addr-exp-mod (lib "typedscm.ss" "typedscm")

(provide lexical-addr-exp?
         la:var-exp? la:quote-exp? la:lambda-exp? la:if-exp? la:app-exp?
         la:var-exp->lexical-addr  la:quote-exp->symbol la:lambda-exp->ids
         la:lambda-exp->body la:app-exp->rator la:app-exp->rands
         la:if-exp->test-exp la:if-exp->true-exp la:if-exp->false-exp
         la:var-exp la:quote-exp la:lambda-exp la:if-exp
         la:app-exp parse-lexical-addr-exp
         ;; all-from lexical-addr-mod.scm
         lexical-addr? lexical-addr
         lexical-addr->name lexical-addr->depth  lexical-addr->position
         parse-lexical-addr)

;; type predicate
(deftype lexical-addr-exp? (type-predicate-for lexical-addr-exp))

;; case testers (discriminators)
(deftype la:var-exp? (-> (lexical-addr-exp) boolean))
(deftype la:quote-exp? (-> (lexical-addr-exp) boolean))
(deftype la:lambda-exp? (-> (lexical-addr-exp) boolean))
(deftype la:if-exp? (-> (lexical-addr-exp) boolean))
(deftype la:app-exp? (-> (lexical-addr-exp) boolean))

;; observers 
(deftype la:var-exp->lexical-addr (-> (lexical-addr-exp) lexical-addr))
(deftype la:quote-exp->symbol (-> (lexical-addr-exp) symbol))
(deftype la:lambda-exp->ids (-> (lexical-addr-exp) (list-of symbol)))
(deftype la:lambda-exp->body (-> (lexical-addr-exp) lexical-addr-exp))
(deftype la:app-exp->rator (-> (lexical-addr-exp) lexical-addr-exp))
(deftype la:app-exp->rands (-> (lexical-addr-exp) (list-of lexical-addr-exp)))
(deftype la:if-exp->test-exp (-> (lexical-addr-exp) lexical-addr-exp))
(deftype la:if-exp->true-exp (-> (lexical-addr-exp) lexical-addr-exp))
(deftype la:if-exp->false-exp (-> (lexical-addr-exp) lexical-addr-exp))

;; constructors
(deftype la:var-exp (-> (lexical-addr) lexical-addr-exp))
(deftype la:quote-exp (-> (symbol) lexical-addr-exp))
(deftype la:lambda-exp (-> ((list-of symbol) lexical-addr-exp) lexical-addr-exp))
(deftype la:if-exp
  (-> (lexical-addr-exp lexical-addr-exp lexical-addr-exp) lexical-addr-exp))
(deftype la:app-exp
  (-> (lexical-addr-exp (list-of lexical-addr-exp)) lexical-addr-exp))

;; parsing/bless input correctness
(deftype parse-lexical-addr-exp (-> (datum) lexical-addr-exp))

(require (lib "all-mod.scm" "lib342") (lib "lexical-addr-mod.scm" "lib342"))

(defrep (lexical-addr-exp datum))

(define lexical-addr-exp?
  (lambda (exp)
    (has-type-trusted
      boolean
      (or (and (la:var-exp? exp)
               (lexical-addr? (la:var-exp->lexical-addr exp)))
          (and (la:quote-exp? exp) 
               (symbol? (la:quote-exp->symbol exp)))
          (and (la:lambda-exp? exp) 
               (lexical-addr-exp? (la:lambda-exp->body exp)))
          (and (la:if-exp? exp)
               (lexical-addr-exp? (la:if-exp->test-exp exp))
               (lexical-addr-exp? (la:if-exp->true-exp exp))
               (lexical-addr-exp? (la:if-exp->false-exp exp)))
          (and (la:app-exp? exp)
               (lexical-addr-exp? (la:app-exp->rator exp))
               (all lexical-addr-exp? (la:app-exp->rands exp)))))))

(define la:var-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents a la:var-exp
    (lexical-addr? exp)))

(define la:quote-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents a la:quote-exp
    (test-type
      boolean?
      (and (list? exp)
           (= (length exp) 2)
           (eq? (car exp) 'quote)
           (symbol? (cadr exp))))))

(define la:lambda-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents a la:lambda-exp
    (test-type
      boolean?
      (and (list? exp)
           (= (length exp) 3)
           (eq? (car exp) 'lambda)
           (let ((ids (cadr exp)))
             (and (list? ids)
                  (all symbol? ids)))))))

(define la:if-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents an la:if-exp
    (test-type
      boolean?
      (and (list? exp)
           (= (length exp) 4)
           (eq? (car exp) 'if)))))

(define la:app-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents an la:app-exp
    (test-type
      boolean?
      (and (list? exp)
           (>= (length exp) 1)
           (or (not (= (length exp) 4))
               (not (eq? ': (cadr exp))))
           (let ((operator (car exp)))
             (and (not (eq? operator 'quote))
                  (not (eq? operator 'lambda))
                  (not (eq? operator 'if))))))))

(define la:var-exp->lexical-addr
  (lambda (vr)
    ;; REQUIRES: vr represents a la:var-exp
    ;; ENSURES: result is the lexical-addr represented by vr
    (if (la:var-exp? vr)
        (test-type lexical-addr? vr)
        (error "la:var-exp->id: not a la:var-exp:" vr))))

(define la:quote-exp->symbol
  (lambda (qe)
    ;; REQUIRES: qe represents a la:quote-exp
    ;; ENSURES: result is the operand in the representation of qe
    (if (la:quote-exp? qe)
        (test-type symbol? (cadr qe))
        (error "la:quote-exp->symbol: not a la:quote-exp:" qe))))

(define la:lambda-exp->ids
  (lambda (lm)
    ;; REQUIRES: lm represents a la:lambda-exp
    ;; ENSURES: result is the list of formals in the representation of lm
    (if (la:lambda-exp? lm)
        (test-type (list-of symbol?) (cadr lm))
        (error "la:lambda-exp->id: not a la:lambda-exp:" lm))))

(define la:lambda-exp->body
  (lambda (lm)
    ;; REQUIRES: lm represents a la:lambda-exp
    ;; ENSURES: result is the body in the representation of lm
    (if (la:lambda-exp? lm)
        (test-type lexical-addr-exp? (caddr lm))
        (error "la:lambda-exp->body: not a la:lambda-exp:" lm))))

(define la:app-exp->rator
  (lambda (ap)
    ;; REQUIRES: ap represents an la:app-exp
    ;; ENSURES: result is the operator in the representation of ap
    (if (la:app-exp? ap)
        (test-type lexical-addr-exp? (car ap))
        (error "la:app-exp->rator: not an la:app-exp:" ap))))

(define la:app-exp->rands
  (lambda (ap)
    ;; REQUIRES: ap represents an la:app-exp
    ;; ENSURES: result is the list of operands in the representation of ap
    (if (la:app-exp? ap)
        (test-type (list-of lexical-addr-exp?) (cdr ap))
        (error "la:app-exp->rands: not an la:app-exp:" ap))))

(define la:if-exp->test-exp
  (lambda (ie)
    ;; REQUIRES: ie represents an la:if-exp
    ;; ENSURES: result is the test expression in ie
    (if (la:if-exp? ie)
        (test-type lexical-addr-exp? (cadr ie))
        (error "la:if-exp->test-exp: not an la:if-exp:" ie))))


(define la:if-exp->true-exp
  (lambda (ie)
    ;; REQUIRES: ie represents an la:if-exp
    ;; ENSURES: result is the true expression in ie
    (if (la:if-exp? ie)
        (test-type lexical-addr-exp? (caddr ie))
        (error "la:if-exp->true-exp: not an la:if-exp:" ie))))

(define la:if-exp->false-exp
  (lambda (ie)
    ;; REQUIRES: ie represents an la:if-exp
    ;; ENSURES: result is the false expression in ie
    (if (la:if-exp? ie)
        (test-type lexical-addr-exp? (cadddr ie))
        (error "la:if-exp->false-exp: not an la:if-exp:" ie))))

(define la:var-exp
  (lambda (la)
    ;; ENSURES: result is a representation of a la:var-exp with lexical address la
    (has-type datum la)))

(define la:quote-exp
  (lambda (sym)
    ;; ENSURES: result is a representation of a la:quote-exp with symbol sym
    (has-type datum (list 'quote sym))))

(define la:lambda-exp
  (lambda (ids body)
    ;; ENSURES: result is a representation of a la:lambda-exp with formals ids
    ;;          and body body
    (has-type datum (list 'lambda ids body))))

(define la:if-exp
  (lambda (test-exp true-exp false-exp)
    ;; ENSURES: result is a representation of an la:if-exp with test
    ;; test, true expression, true-exp, and false expression false-exp
    (has-type datum (list 'if test-exp true-exp false-exp))))

(define la:app-exp
  (lambda (rator rands)
    ;; ENSURES: result is a representation of an la:app-exp with operator rator
    ;;          and operands rands
    (has-type datum (cons rator rands))))
  
(define parse-lexical-addr-exp
  (lambda (exp)
    (test-type
      datum?
      (cond
       ((lexical-addr? exp) (la:var-exp exp))
       ((la:quote-exp? exp) (la:quote-exp (cadr exp)))
       ((la:lambda-exp? exp)
        (la:lambda-exp (cadr exp) (parse-lexical-addr-exp (caddr exp))))
       ((la:if-exp? exp) (la:if-exp (parse-lexical-addr-exp (cadr exp))
                                    (parse-lexical-addr-exp (caddr exp))
                                    (parse-lexical-addr-exp (cadddr exp))))
       ((la:app-exp? exp) (la:app-exp (parse-lexical-addr-exp (car exp))
                                      (map parse-lexical-addr-exp (cdr exp))))
       (else (error "parse-lexical-addr-exp: bad syntax in expression: "
                    exp))))))

) ;; end module
