;;; $Id: slib-hashtab.scm,v 1.9 2006/04/25 17:32:42 leavens Exp $
; "hashtab.scm", hash tables for Scheme.
; Copyright (c) 1992, 1993, 2003 Aubrey Jaffer
;
;Permission to copy this software, to modify it, to redistribute it,
;to distribute modified versions, and to use it for any purpose is
;granted, subject to the following restrictions and understandings.
;
;1.  Any copy made of this software must include this copyright notice
;in full.
;
;2.  I have made no warranty or representation that the operation of
;this software will be error-free, and I am under no obligation to
;provide any services, by way of maintenance, update, or otherwise.
;
;3.  In conjunction with products arising from the use of this
;material, there shall be no use of my name in any advertising,
;promotional, or sales literature without prior written consent in
;each case.

(module slib-hashtab (lib "typedscm.ss" "typedscm")

(provide slib:hash-table? predicate->hash make-hash-table predicate->hash-asso
         hash-inquirer hash-associator hash-remover hash-map
         hash-for-each hash-rehasher)

;;@code{(require 'hash-table)}
;;@ftindex hash-table

(deftype slib:hash-table? (forall (K V) (type-predicate-for (slib:hash-table K V))))
(deftype predicate->hash
  (forall (T) (-> ((-> (T T) boolean)) (-> (datum number) number))))
(deftype make-hash-table
  (forall (K V) (-> (number) (slib:hash-table K V))))
(deftype predicate->hash-asso
  (forall (K V T) (-> ((-> (T T) boolean)) (-> (K (slib:hash-table K V)) (pair-of K V)))))
(deftype hash-inquirer
  (forall (K V T) (-> ((-> (T T) boolean)) (-> ((slib:hash-table K V) K) datum))))
(deftype hash-associator
  (forall (K V T) (-> ((-> (T T) boolean))
		      (-> ((slib:hash-table K V) K V) (slib:hash-table K V)))))
(deftype hash-remover
  (forall (K V T) (-> ((-> (T T) boolean))
		      (-> ((slib:hash-table K V) K) (slib:hash-table K V)))))
(deftype hash-map
  (forall (K V W) (-> ((-> (K V) W) (slib:hash-table K V))
		      (slib:hash-table K W))))
(deftype hash-for-each
  (forall (K V) (-> ((-> (K V) void) (slib:hash-table K V))
		    void)))
(deftype hash-rehasher
  (forall (K V T) (-> ((-> (T T) boolean))
		    (-> ((slib:hash-table K V) number)
			(slib:hash-table K V)))))

(require (lib "slib-hash.scm" "typedscm")
         (lib "slib-alist.scm" "typedscm"))

(defrep (forall (K V) (slib:hash-table K V)
		      (vector-of (list-of (pair-of K V)))))

(define slib:hash-table?
  (has-type-trusted
   (forall (K V) (type-predicate-for (slib:hash-table K V)))
   vector?))

;;@body
;;Returns a hash function (like @code{hashq}, @code{hashv}, or
;;@code{hash}) corresponding to the equality predicate @var{pred}.
;;@var{pred} should be @code{eq?}, @code{eqv?}, @code{equal?}, @code{=},
;;@code{char=?}, @code{char-ci=?}, @code{string=?}, or
;;@code{string-ci=?}.
(define (predicate->hash pred)
  (has-type-trusted
    (-> (datum number) number)
    (cond ((eq? pred eq?) hashq)
	  ((eq? pred eqv?) hashv)
	  ((eq? pred equal?) hash)
	  ((eq? pred =) hashv)
	  ((eq? pred char=?) hashv)
	  ((eq? pred char-ci=?) hashv)
	  ((eq? pred string=?) hash)
	  ((eq? pred string-ci=?) hash)
	  (else (error "unknown predicate for hash" pred)))))

;;@noindent
;;A hash table is a vector of association lists.

;;@body
;;Returns a vector of @var{k} empty (association) lists.
(define (make-hash-table k)
    (make-vector k (has-type (forall (KT VT) (list-of (pair-of KT VT))) '())))

;;@noindent
;;Hash table functions provide utilities for an associative database.
;;These functions take an equality predicate, @var{pred}, as an argument.
;;@var{pred} should be @code{eq?}, @code{eqv?}, @code{equal?}, @code{=},
;;@code{char=?}, @code{char-ci=?}, @code{string=?}, or
;;@code{string-ci=?}.

;;@body
;;Returns a hash association function of 2 arguments, @var{key} and
;;@var{hashtab}, corresponding to @var{pred}.  The returned function
;;returns a key-value pair whose key is @var{pred}-equal to its first
;;argument or @code{#f} if no key in @var{hashtab} is @var{pred}-equal to
;;the first argument.
(define (predicate->hash-asso pred)
  (let ((hashfun (predicate->hash pred))
	(asso
	 (has-type-trusted
	   (forall (s t) (-> (s (list-of (pair-of s t))) (pair-of s t)))
	   (predicate->asso pred))))
    (lambda (key hashtab)
      (asso key
	    (vector-ref hashtab (hashfun key (vector-length hashtab)))))))
;;@body
;;Returns a procedure of 2 arguments, @var{hashtab} and @var{key}, which
;;returns the value associated with @var{key} in @var{hashtab} or
;;@code{#f} if @var{key} does not appear in @var{hashtab}.
(define (hash-inquirer pred)
  (let ((hashfun (predicate->hash pred))
	(ainq
	 (has-type-trusted
	   (forall (s t) (-> ((list-of (pair-of t s)) t) datum))
	   (alist-inquirer pred))))
    (lambda (hashtab key)
      (ainq (vector-ref hashtab (hashfun key (vector-length hashtab)))
	    key))))
;;@body
;;Returns a procedure of 3 arguments, @var{hashtab}, @var{key}, and
;;@var{value}, which modifies @var{hashtab} so that @var{key} and
;;@var{value} associated.  Any previous value associated with @var{key}
;;will be lost.
(define (hash-associator pred)
  (let ((hashfun (predicate->hash pred))
	(asso
	 (has-type-trusted
	   (forall (k v)
	     (-> ((list-of (pair-of k v)) k v) (list-of (pair-of k v))))
	   (alist-associator pred))))
    (lambda (hashtab key val)
      (let* ((num (hashfun key (vector-length hashtab))))
	(vector-set! hashtab num
		     (asso (vector-ref hashtab num) key val)))
      hashtab)))
;;@body
;;Returns a procedure of 2 arguments, @var{hashtab} and @var{key}, which
;;modifies @var{hashtab} so that the association whose key is @var{key} is
;;removed.
(define (hash-remover pred)
  (let ((hashfun (predicate->hash pred))
	(arem
	 (has-type-trusted
	   (forall (k v)
	     (-> ((list-of (pair-of k v)) s) (list-of (pair-of k v))))
	 (alist-remover pred))))
    (lambda (hashtab key)
      (let* ((num (hashfun key (vector-length hashtab))))
	(vector-set! hashtab num
		     (arem (vector-ref hashtab num) key)))
      hashtab)))
;;@args proc hash-table
;;Returns a new hash table formed by mapping @var{proc} over the
;;keys and values of @var{hash-table}.  @var{proc} must be a function of 2
;;arguments which returns the new value part.
(define (hash-map proc ht)
  (let ((nht (make-vector (vector-length ht))))
    (letrec ((iter
	      (lambda (i)
		(if (negative? i)
		    nht
		    (begin
		      (vector-set! nht i (alist-map proc (vector-ref ht i)))
		      (iter (+ -1 i)))))))
      (iter (+ -1 (vector-length ht))))))

;;@args proc hash-table
;;Applies @var{proc} to each pair of keys and values of @var{hash-table}.
;;@var{proc} must be a function of 2 arguments.  The returned value is
;;unspecified.
(define (hash-for-each proc ht)
  (letrec ((iter
	    (lambda (i)
	      (if (not (negative? i))
		  (begin
		    (alist-for-each proc (vector-ref ht i))
		    (iter (+ -1 i)))))))
    (iter (+ -1 (vector-length ht)))))
    
;;@body
;;@0 accepts a hash table predicate and returns a function of two
;;arguments @var{hashtab} and @var{new-k} which is specialized for
;;that predicate.
;;
;;This function is used for nondestructively resizing a hash table.
;;@var{hashtab} should be an existing hash-table using @1, @var{new-k}
;;is the size of a new hash table to be returned.  The new hash table
;;will have all of the associations of the old hash table.
(define (hash-rehasher pred)
  (let ((hashfun (predicate->hash pred)))
    (lambda (hashtab newk)
      (let ((newtab (make-hash-table newk)))
	(hash-for-each
	 (lambda (key value)
	   (let ((num (hashfun key newk)))
	     (has-type-trusted
	       void
	       (vector-set! newtab num
			    (cons (cons key value)
				  (vector-ref newtab num))))))
	 hashtab)
	newtab))))

) ;; end module
