;;; $Id: tc-scheme-annotate.scm,v 1.93 2008/02/22 04:19:56 dorn Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; This file contains rules for annotating Scheme syntax with typing
;;; information.
;;;
;;; AUTHOR: Brian Dorn with help from Gary T. Leavens

(module tc-scheme-annotate (lib "typedscm.ss" "typedscm")

(provide tc:scheme-top-level-annotate
	 tc:scheme-annotate
	 tc:annotation->type)

(require-for-syntax mzscheme)
(require (lib "tc-scheme-parser.scm" "typedscm")
         (lib "tc-scheme-abstract-syntax.scm" "typedscm")
         (lib "tc-scheme-unparser.scm" "typedscm")
	 (lib "tc-util.scm" "typedscm")
	 (lib "tc-tracing.scm" "typedscm")
	 (lib "maybe.scm" "typedscm")
	 (lib "tc-types.scm" "typedscm")
	 (lib "tc-global-var-types.scm" "typedscm")
	 (lib "tc-type-translate.scm" "typedscm")
	 (lib "tc-type-helpers.scm" "typedscm"))

(define tc:scheme-top-level-annotate
  (lambda (pi e)
    (tc:set-current-subst-list! (list (tc:type-expr-null-subst)))
    (tc:scheme-annotate pi e)))

(deftype tc:scheme-annotate
  (forall (syn)
    (-> (tc:type-env syn)
	(tc:atree-of syn))))
(define tc:scheme-annotate
  (lambda (pi e)
    ;; ENSURES: result is the annotation of the given syntax, e.
    ;; This procedure is essentially used to simulate static overloading
    ;; in Scheme.
    (let ((annotate
	   (cond
	    ((tc:scheme-form? e) tc:scheme-form-annotate)
	    ((tc:scheme-exp? e) tc:scheme-exp-annotate)
	    ((tc:scheme-datum? e) tc:scheme-datum-annotate)
	    ((tc:scheme-formals? e) tc:scheme-formals-annotate)
	    ((tc:scheme-body? e) tc:scheme-body-record-annotate)
	    ((tc:scheme-definition? e) tc:scheme-definition-annotate)
	    ((tc:variant-decl? e) tc:variant-decl-record-annotate)
	    ((tc:field-pred-binding? e) tc:field-pred-binding-record-annotate)
	    ((tc:scheme-cond-clause? e) tc:scheme-cond-clause-annotate)
	    ((tc:scheme-case-clause? e) tc:scheme-case-clause-annotate)
	    ((tc:scheme-binding-spec? e) tc:scheme-binding-spec-annotate)
	    ((tc:scheme-cases-clause? e) tc:scheme-cases-clause-annotate)
	    ((tc:scheme-iteration-spec? e) tc:scheme-iteration-spec-annotate)
	    ((tc:scheme-program? e) tc:scheme-program-annotate)
	    ((tc:input-type-exp? e) tc:input-type-exp-annotate)
	    ((tc:scheme-module-name? e) tc:scheme-module-name-annotate)
	    ;; more cases go here	    
	    (else
	     (error "tc:scheme-annotate: Unable to annotate syntax type of: "
		    (tc:unparse-scheme-syntax e)))
	    )))
      (if (> (type-tracing-level) 3)
	  (begin
	    (displayln "ANNOTATE called with: ")
	    (pretty-print (tc:unparse-scheme-syntax e))))
      (let ((annotate-result (annotate pi e)))
	(cond
	 ((> (type-tracing-level) 5)
	  (displayln "ANNOTATE returned: ")
	  (pretty-print annotate-result))
	 ((> (type-tracing-level) 3)
	  (displayln "ANNOTATE returned")))
	annotate-result))))

;; Local definitions of the rule helpers that partially parameterize
;; them with the method dictionary tc:type-expr-as-unifiable
;; and the tc:scheme-annotate procedure
(define tc:axiom (tc:axiom-helper (tc:type-expr-as-unifiable)))

(define tc:rule
  (lambda (hs conc)
    ((tc:rule-helper (tc:type-expr-as-unifiable)
		    tc:scheme-annotate)
     hs conc)))
(define tc:rule-if
  (lambda (hs side-cond side-defs conc)
    ((tc:rule-if-helper (tc:type-expr-as-unifiable)
			tc:scheme-annotate)
    hs side-cond side-defs conc)))
(define tc:rule-seq-local
  (lambda (hs conc)
    ((tc:rule-seq-helper (tc:type-expr-as-unifiable)
			 tc:scheme-annotate)
     hs conc)))
(define tc:rule-or-local
  (lambda rules
    (apply
     (tc:rule-or-helper (tc:type-expr-as-unifiable)
			tc:scheme-annotate)
     rules)))
  
(define-syntax tc:rule-seq
  (syntax-rules (list)
    ((tc:rule-seq (list h1 ...) conc)
     (tc:rule-seq-local (list (delay h1) ...) conc))))
(define-syntax tc:rule-or
  (syntax-rules ()
    ((tc:rule-or t1 ...)
     (tc:rule-or-local (delay t1) ...))))


;; The void type
(define *tc:void* (tc:basic-type-expr 'void))
(define *tc:boolean* (tc:basic-type-expr 'boolean))
(define *tc:datum* (tc:basic-type-expr 'datum))
(define *tc:symbol* (tc:basic-type-expr 'symbol))

;; programs 
(deftype tc:scheme-program-annotate
  (forall (syn)
    (-> (tc:type-env syn)
        (tc:atree-of syn))))
(define tc:scheme-program-annotate
  (lambda (pi prog)
    (cases tc:scheme-program prog
	   (tc:scheme-program-normal
	    (position require-forms deftypes forms)

;;;	    (displayln "Forms: " forms)
;;;	    (displayln "Requires: " (tc:scheme-forms->requires forms))
	    
	    (let* ((require-forms (append require-forms (tc:scheme-forms->requires forms)))
		   (deftypes (append deftypes (tc:scheme-forms->deftypes forms)))
		   (req-types (tc:listof-new-variable-type-expr (length require-forms)))
		   (deft-types (tc:listof-new-variable-type-expr (length deftypes)))
		   (post-deft-env-var (tc:new-logicalvar))		   
		   (bindings (tc:scheme-forms->bindings forms #f '()))
		   (commands (tc:scheme-forms->commands forms))
		   (varrefs
		    (map tc:scheme-binding-spec->varref bindings))
		   (vars
		    (map tc:varref->variable varrefs))
		   ;; Build the new environment to typecheck the forms.  We bind
		   ;; all vars to type variables...
		   (augmented-env (tc:env-join
				   pi                 ;;[[CHECKME Should the take precendence?]]
				   (tc:env-update*
				    vars
				    (map (lambda (x)
					   (tc:bind-type-expr
					    'letrec
					    (tc:new-variable-type-expr)))
					 vars)
				    (tc:env-empty))))
		   ;;(form-types (tc:listof-new-variable-type-expr (length forms)))
		   (binding-types (tc:listof-new-variable-type-expr (length bindings)))
		   (reqs-env-var (tc:new-logicalvar))
		   (env-var (tc:new-logicalvar)))
	      
	      (if (> (type-tracing-level) 4)
		  (tc:env-for-each
		   (lambda (k v)
		     (displayln "Key: " k)
		     (displayln "Bind Type: " (tc:binding->binder v))
		     (displayln "Val: " (tc:binding->type-expr v)))
		   augmented-env))	       

	      ;; The rule for normal programs
	      (tc:rule-seq-local
	       (append
		;; Infer on requires first
		(map (lambda (req type)
		       (delay (tc:hyp (:- pi (: req type)))))
		     require-forms req-types)
		(list
		 (delay (tc:def
			 (lambda (ts)
			   (let ((envs (cons pi (map tc:declaration-type-expr->env ts))))
			     (tc:type-expr-bind
			      reqs-env-var
			      (tc:declaration-type-expr
			       (tc:env-join* envs))))))))
		;; Infer the types on the deftypes
		(map (lambda (deft type)
		       (delay (tc:hyp (:?- reqs-env-var (: deft type)))))
		     deftypes deft-types)
		(list
		 ;; Create a new environment where deftypes are bound and any non
		 ;; bound identifiers are set to type vars
		 (delay (tc:def
			 (lambda (ts)
			   ;; Augmented-env should take the least precedence here so deftyped
			   ;; things overwrite the type variables
			   (let* ((envs (cons augmented-env
					      (map tc:declaration-type-expr->env ts))))
			     (tc:type-expr-bind
			      post-deft-env-var
			      (tc:declaration-type-expr
			       (tc:env-join*
				(reverse envs)))))))))		

		(map (lambda (binding type)
		       (delay (tc:hyp (:?- post-deft-env-var (: binding type)))))
		     bindings binding-types)
		;; The basic commands must be void	
		(map (lambda (command)
		       (delay (tc:hyp (:?- post-deft-env-var (: command *tc:void*)))))
		     commands)
		;; Build our new environment based on the results, this ts arg contains
		;; both the deftypes and types from other defines.
		(list		  
		 (delay (tc:def
			 (lambda (ts)
			   (let* ((decl-ts (tc:filter tc:declaration-type-expr? ts))
				  (envs (map tc:declaration-type-expr->env decl-ts)))
			     (tc:type-expr-bind
			      env-var
			      (tc:declaration-type-expr
			       (tc:env-join* envs))))))))) 
	       ;;------------------------------------
	       (:- pi (: prog (tc:variable-type-expr env-var))))))
	     (else (error "tc:scheme-program-annotate cannot handle: "
			  (tc:unparse-scheme-syntax prog))))))

;; forms
(define tc:scheme-form-annotate
  (lambda (pi form)
    (cases tc:scheme-form form
	   (tc:command
	    (position exp)
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: exp t)))
		       ;;--------------------
		       (:- pi (: form t)))))
	   (tc:load-command
	    (position operator filename program)
	    (displayln "Type checking \"" filename "\" ...")
	    (let ((t (tc:new-variable-type-expr)))
	      (let ((load-result
		     (tc:rule (list
			       (:- pi (: program t)))
			      ;;-----------------------
			      (:- pi (: form t)))))
		(displayln "... done type checking \"" filename "\"")
		load-result)))
	   (tc:definition
	    (position definition)
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: definition t)))
		       ;;--------------------
		       (:- pi (: form t)))))
	   (tc:syntax-definition
	    (position spec)
	    ;; These are ignored
	    (tc:axiom (:- pi (: form *tc:void*))))

	   ;;[[[FINISHME]]]	  
	   ;;	   (tc:begin-form
	   ;; SEE Load now?
	   ;;	    (position forms)   
	   ;;	    )
	   
	   ;; for the type checker
	   (tc:module
	    (position identifier initial-required-module-name
		      provides requires-for-syntax requires defreps deftypes definitions)
	    (let ((mod-exp (tc:module-exp position identifier initial-required-module-name
					  provides requires-for-syntax requires defreps
					  deftypes definitions))
		  (mod-type (tc:new-variable-type-expr))
		  (result-env-var (tc:new-logicalvar)))
	    (tc:rule-seq-local
		(list
		 (delay (tc:hyp (:- pi (: mod-exp mod-type))))
		 (delay (tc:def
			 (lambda (ts)
			   (let ((env (tc:declaration-type-expr->env (car ts))))
			     (tc:type-expr-bind
			      result-env-var
			      (tc:declaration-type-expr
			       ;; Filter based on the provides
			       (tc:env-filter
				(lambda (k v)
				  (memq k provides))
				;; We return back to the abstract types by NOT reversing
				env))))))))
		;;------------------------------------------------------------
	       (:- pi (: form (tc:declaration-type-expr
			       (tc:env-update
				identifier
				(tc:bind-type-expr
				 'lambda
				 (tc:variable-type-expr result-env-var))
				(tc:env-empty))))))))
		      
	   (tc:require-form
	    (position module-names)
	    (let ((module-types (tc:listof-new-variable-type-expr (length module-names)))
		  (env-var (tc:new-logicalvar)))
	      (tc:rule-seq-local
	       (append
		;; Infer the types on each of the modules listed
		(map (lambda (module-name type)
		       (delay (tc:hyp (:- pi (: module-name type)))))
		     module-names module-types)
		;; Merge all the envs together
		(list (delay (tc:def
			      (lambda (ts)
				(let ((envs (map tc:declaration-type-expr->env ts)))
				  (tc:type-expr-bind
				   env-var
				   (tc:declaration-type-expr
				    ;; [[[CHECKME What about name clashes?]]]
				    (tc:env-join* envs)))))))))
	       ;;--------------------------------------------------
	       (:- pi (: form (tc:variable-type-expr env-var))))))

	   (else
	    (error "Cannot handle this type of form yet: "
		   (tc:unparse-scheme-syntax form))))))

;; Do modular type checking by not type checking modules required
;; from other modules (when true)
(deftype *tc:trust-deftypes-in-modules* boolean)
(define *tc:trust-deftypes-in-modules* #t)

;; module names
(deftype tc:scheme-module-name-annotate
  (-> (tc:type-env tc:scheme-module-name)
      (tc:atree-of tc:scheme-module-name)))
(define tc:scheme-module-name-annotate
  (lambda (pi modname)
    (let ((annotate-file
	   (lambda (position path)
	     (let ((mod-sym (tc:extract-name path))
		   (prog-exp (tc:parse-scm-file path position))
		   (file-type (tc:new-variable-type-expr))
		   (mod-env-var (tc:new-logicalvar)))
	       (if (= 0 (tc:position->end-line
			 (tc:scheme-program->position prog-exp)))
		   ;; file not found
		   (tc:axiom
		    (:- pi (: modname (tc:bad-file-error-maker position path))))
		   (tc:rule-seq
		    (list
		     (tc:hyp (:- (tc:env-empty) (: prog-exp file-type)))
		     (tc:def
		      (lambda (ts)
			(let* ((file-env
				(tc:declaration-type-expr->env (car ts)))
			       (mod-env-type-expr
				(tc:binding->type-expr
				 (tc:env-value mod-sym file-env))))
			  (tc:type-expr-bind mod-env-var
					     mod-env-type-expr)))))
		    ;;-----------------------------------------------
		    (:- pi (: modname (tc:variable-type-expr mod-env-var)))))))))
      (cases tc:scheme-module-name modname
	     (tc:module-id
	      (position identifier)	      	      
	      (if (and (tc:env-bound? identifier pi)
		       (tc:declaration-type-expr?
			(tc:binding->type-expr (tc:env-value identifier pi))))
		  ;; The module has been loaded in environment
		  (tc:axiom (:- pi (: modname (tc:binding->type-expr
					       (tc:env-value identifier pi)))))
		  ;; Module not found
		  (tc:axiom
		   (:- pi (: modname
			     (tc:unknown-module-error-maker modname identifier))))))
	     (tc:unix-relative-path-string
	      (position path)
	      (annotate-file position path))
	     (tc:file
	      (position path)
	      (annotate-file position path))
	     (tc:lib
	      (position filename collects)
	      (annotate-file position
			     (tc:scheme-module-name->file-location modname)))
	     (tc:import-from-module-name
	      (name)
	      (if *tc:trust-deftypes-in-modules*
		  (tc:extract-deftypes-from-file
		   pi modname (tc:scheme-module-name->file-location name))
		  (tc:scheme-module-name-annotate pi name)))
	     (else
	      (error
	       "tc:scheme-module-name-annotate: missing case for:" modname))
	     ))))

;; For required modules, we trust their deftypes and just make sure we have deftypes for all provides
(deftype tc:extract-deftypes-from-file
  (-> (tc:type-env tc:scheme-module-name tc:position string)
      (tc:atree-of tc:scheme-module-name)))
(define tc:extract-deftypes-from-file
  (lambda (pi modname path)
    (let* ((mod-sym (tc:extract-name path))
	   (position (tc:scheme-module-name->position modname))
	   (prog-exp (tc:parse-scm-file path position))
	   (file-type (tc:new-variable-type-expr))
	   (mod-env-var (tc:new-logicalvar)))
      (if (= 0 (tc:position->end-line
		(tc:scheme-program->position prog-exp)))
	  ;; file not found
	  (tc:axiom
	   (:- pi (: modname (tc:bad-file-error-maker position path))))
	  (let* ((module-form (car (tc:scheme-program->forms prog-exp))))
	    (cases tc:scheme-form module-form
	       (tc:module (position identifier intial-required-module-name
				    provides requires-for-syntax requires defreps deftypes definitions)
			  (tc:rule-seq-local
			   (append
			    ;; Build the list of names bound with deftypes here
			    (map (lambda (a-deftype t)
				   (delay (tc:hyp (:- (tc:env-empty) (: a-deftype t)))))
				 deftypes (tc:listof-new-variable-type-expr (length deftypes)))
			    ;; Grab definitions from required modules
			    (map (lambda (required-name t)
				   (delay (tc:hyp (:- (tc:env-empty) (: required-name t)))))
				 requires (tc:listof-new-variable-type-expr (length requires)))
			    (list
			     ;; Make sure that there is a deftype for every provided symbol
			     (delay (tc:when
				     (lambda (ts)
				       (= (length provides)
					  (length
					   (tc:intersect
					    provides
					    ;;All of the names for which there is a deftype
					    (apply append 
						   (map (lambda (t) (tc:env-names (tc:declaration-type-expr->env t))) ts))))))))
			     (delay (tc:def
				     (lambda (ts)
				       (let* ((file-env
					       (tc:env-filter
						(lambda (k v)
						  (memq k provides))
						(tc:env-join* (map tc:declaration-type-expr->env ts))))
					      (mod-env-type-expr
					       (tc:declaration-type-expr file-env)))
					 (tc:type-expr-bind mod-env-var
							    mod-env-type-expr)))))))
			   ;;-----------------------------------------------
			   (:- pi (: modname (tc:variable-type-expr mod-env-var)))))
	       (else (error "tc:extract-deftypes-from-file: a non-module form was provided: " (tc:unparse-scheme-syntax prog-exp)))
	       ))))))

;; expressions
(deftype tc:scheme-exp-annotate
  (-> (tc:type-env tc:scheme-exp)
	(tc:atree-of tc:scheme-exp)))
(define tc:scheme-exp-annotate
  (lambda (pi e)
    (cases tc:scheme-exp e
	   (tc:varref
	    (position variable)
	    ;; [[[FIXME, turn this into a rule-or and rule-ifs]]]	    
	    (if (and (tc:env-bound? variable pi)
		     ;; disallow resolution of module names
		     (not (tc:declaration-type-expr? (tc:binding->type-expr
						      (tc:env-value variable pi)))))
		(let ((binding (tc:env-value variable pi)))
		  (let ((type (tc:binding->type-expr binding)))
		    (if (tc:bound? 'let binding)			
			(let ((new-type (tc:type-expr-instance (tc:subst-bindings pi type)))) 
			  ;; TC:AXIOM: Let-bound variable
			  (tc:axiom (:- pi (: e new-type))))
			;; TC:AXIOM: Non let-bound variable
			(tc:axiom (:- pi (: e type))))))
		(if (tc:env-bound? variable (tc:global-var-types))
		    (let ((global-type (tc:type-expr-instance					
					(tc:env-value variable (tc:global-var-types))))) 	     
		      (tc:axiom (:- pi (: e global-type))))
		    ;; TC:AXIOM: Unknown variable
		    (tc:axiom (:- pi (: e (tc:varref-error-maker e)))))))
	   (tc:quotation (position datum)
			 (let ((t (tc:new-variable-type-expr)))
			   (tc:rule (list
				     (:- pi (: datum t)))
				    ;;----------------------
				    (:- pi (: e t)))))
	   (tc:self-evaluating
	    (position datum)
	    (let ((t (tc:infer-simple-datum-type datum)))
	      (tc:axiom (:- pi (: e t)))))
	   (tc:procedure-call
	    (position operator operands)
	    (let* ((rt (tc:new-variable-type-expr))
		   (operand-ts (tc:listof-new-variable-type-expr (length operands)))
		   (oper-type (tc:new-variable-type-expr)))
	      (tc:rule (append
			(list
			 (:- pi (: operator oper-type)))
			(map
			 (lambda (operand t)
			   (:- pi (: operand t)))
			 operands operand-ts)
			(list
			 (<: oper-type (tc:function-type-expr operand-ts rt))))
		       ;;----------------------------
		       (:- pi (: e rt)))))
	   (tc:lambda-exp
	    (position formals body)
	    (let ((formals-var (tc:new-variable-type-expr))
		  (rt (tc:new-variable-type-expr))
		  (formals+pi-var (tc:new-logicalvar))
		  (formals-type-list '()))	      
	      (tc:rule-seq (list
			    (tc:hyp (:- pi (: formals formals-var)))	       
			    (tc:def ;; Extend environment with formals
			     (lambda (ts)
			       (let ((formals-env (tc:declaration-type-expr->env (car ts))))
				 (let ((formals+pi-env
					(tc:env-join formals-env pi)))
				   (begin
				     (set! formals-type-list (tc:lookup-formals formals formals-env))
				     (tc:type-expr-bind formals+pi-var
							(tc:declaration-type-expr formals+pi-env)))))))
			    (tc:hyp (:?- formals+pi-var (: body rt))))
			   ;;------------------------------
			   (:- pi (: e (tc:function-type-expr
					formals-type-list
					rt))))))
	   (tc:named-lambda-exp
	    (position variable formals body)
	    ;; If the name is bound and not letrec-bound, we'll do top down
	    (if (and (tc:env-bound? variable pi)
		     (not (eq? 'letrec
			       (tc:binding->binder
				(tc:env-value variable pi)))))
		(let ((defined-type
			(tc:binding->type-expr (tc:env-value variable pi))))
		  ;; Verify that this can work with the given type,
		  ;; which must be function type 
		  ;; and have the correct number of arguments
		  (cond
		   ((and (tc:function-type-expr? defined-type)
			 (= (let ((def-length
				    (length
				     (tc:function-type-expr->arg-types
				      defined-type))))
			      (if (tc:varargs-function-type-expr? defined-type)
				  (- def-length 1)
				  def-length))
			    (length (tc:scheme-formals->names formals))))
		    (let* ((arg-types
			    (let ((ts (tc:function-type-expr->arg-types
				       defined-type)))
			      (if (tc:varargs-function-type-expr? defined-type)
				  (let ((temp (cdr (reverse ts))))
				    (reverse
				     (cons (tc:applied-type-expr
					    (tc:basic-type-expr 'list-of)
					    (list (car temp)))
					   (cdr temp))))
				  ts)))
			   (formals+pi-env
			    (tc:env-update*
			     (tc:scheme-formals->names formals)
			     (map (lambda (type)
				    (tc:bind-type-expr 'lambda type))
				  arg-types)
			     pi))
			   (rt (tc:new-variable-type-expr))
			   (rt2 (tc:new-variable-type-expr))
			   (declared-rt
			    (tc:function-type-expr->result-type defined-type)))
		      (tc:rule-or
		       ;; this first special case may be more efficient,
		       ;; most of the time.
		       (tc:rule (list
				 (:- formals+pi-env (: body declared-rt)))
				;;------------------------
				(:- pi (: e defined-type)))
		       ;; this is the general rule
		       (tc:rule (list
				 (:- formals+pi-env (: body rt))
				 (<: rt declared-rt))
				;;------------------------
				(:- pi (: e defined-type)))
		       ;; Nice error messages are generated
		       ;; by trying the declaration again
		       (tc:rule (list
				 (:- formals+pi-env (: body declared-rt)))
				;;------------------------
				(:- pi (: e defined-type)))
		       )))
		   ((and (tc:type-predicate-for-type-expr? defined-type)
			 (= 1 (length (tc:scheme-formals->names formals))))
		    (let* ((formals-names (tc:scheme-formals->names formals))
			   (formals-ts
			    (tc:listof-new-variable-type-expr
			     (length formals-names)))
			   (formals+pi-env
			    (tc:env-update*
			     formals-names
			     (map (lambda (type)
				    (tc:bind-type-expr 'lambda type))
				  formals-ts)
			     pi)))
		      (tc:rule (list
				(:- formals+pi-env (: body *tc:boolean*)))
			       ;;----------------------------
			       (:- pi (: e defined-type)))))
		   (else	      
		    ;; Declared type does not match given type
		    (tc:axiom
		     (:- pi (: e (tc:defined-lambda-mismatch-error-maker
				  e defined-type)))))))
		;; When not defined, treat this as a normal lambda
		(let ((lambda-exp (tc:lambda-exp position formals body))
		      (t (tc:new-variable-type-expr)))
		  (tc:rule (list (:- pi (: lambda-exp t)))
			   ;;-----------------------------
			   (:- pi (: e t))))))
	   (tc:conditional-exp
	    (position test consequent alternate)
	    (let* ((t1 (tc:new-variable-type-expr))
		   (t2 (tc:new-variable-type-expr))
		   (supertype-var (tc:new-logicalvar))
		   (subtype-var (tc:new-logicalvar))
		   (supertype (tc:variable-type-expr supertype-var)))
	      (tc:rule (list
			(:- pi (: test *tc:boolean*)) 
			(:- pi (: consequent t1))
			(:- pi (: alternate t2))
			(<:> t2 t1 subtype-var supertype-var))
		       ;; -------------------------
		       (:- pi (: e supertype)))))
	   (tc:conditional-stmt
	    (position test consequent)
	    (tc:rule (list
		      (:- pi (: test *tc:boolean*))
		      (:- pi (: consequent *tc:void*)))
		     ;; ----------------------
		     (:- pi (: e *tc:void*))))
	   (tc:assignment
	    ;;[[CHECKME: what about polymorphic vars?]]
	    (position variable exp)
	    (let ((expected-type (tc:new-variable-type-expr))
		  (inferred-type (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: (tc:varref position variable) expected-type))
			(:- pi (: exp inferred-type))
			(<: inferred-type expected-type))
		       ;;--------------------
		       (:- pi (: e *tc:void*)))))
	   (tc:macro-use
	    (position keyword datums)
	    ;; we ignore this, but it could have any type
	    (tc:axiom (:- pi (: e (tc:new-variable-type-expr)))))
	   (tc:let-syntax-exp
	    (position syntax-specs body)
	    ;; Just pass on body result
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: body t)))
		       ;;-----------------------
		       (:- pi (: e t)))))
	   (tc:letrec-syntax-exp
	    (position syntax-specs body)
	    ;; Just pass on body result
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: body t)))
		       ;;-----------------------
		       (:- pi (: e t)))))
	   ;; derived expressions
	   (tc:cond-exp
	    (position clauses)
	    (let ((clause-var-types (tc:listof-new-variable-type-expr (length clauses))))
	      (let ((generate-result (tc:generate-subtype-rules clause-var-types)))
		(let ((subtype-rules (car generate-result))
		      (supertype (tc:variable-type-expr (cddr generate-result))))
		  (tc:rule (append
			    (map (lambda (clause t)
				   (:- pi (: clause t)))
				 clauses clause-var-types)
			    ;; Generated subtype rules
			    subtype-rules)
			   ;;---------------------
			   (:- pi (: e supertype)))))))	   
	   (tc:cond-stmt
	    (position clauses)
	    (tc:rule (map (lambda (clause)
			    (:- pi (: clause *tc:void*)))
			  clauses)
		     ;;--------------------
		     (:- pi (: e *tc:void*))))
	   (tc:case-exp
	    (position exp clauses)
	    (let ((exp-typ (tc:new-variable-type-expr))
		  (key-env-var (tc:new-logicalvar))
		  (clause-var-types (tc:listof-new-variable-type-expr (length clauses))))
	      (let ((generate-result (tc:generate-subtype-rules clause-var-types)))
		(let ((subtype-rules (car generate-result))
		      (supertype (tc:variable-type-expr (cddr generate-result))))
		  (tc:rule-seq-local (append
				      (list
				       (delay (tc:hyp (:- pi (: exp exp-typ))))
				       (delay (tc:def (lambda (ts)
							(tc:type-expr-bind
							 key-env-var
							 (tc:declaration-type-expr
							  (tc:env-update
							   '*tc:case-key* (tc:bind-type-expr 'lambda (car ts))
							   pi)))))))
				      (map (lambda (clause t)
					     (delay (tc:hyp (:?- key-env-var (: clause t)))))
					   clauses clause-var-types)
				
				      (map (lambda (sub-rule)
					     (delay (tc:hyp sub-rule)))
					   subtype-rules))
				     ;;--------------------------------
				     (:- pi (: e supertype)))))))
	   (tc:case-stmt
	    (position exp clauses)
	    (let ((exp-typ (tc:new-variable-type-expr))
		  (key-env-var (tc:new-logicalvar)))
	      (tc:rule-seq-local (append
				  (list
				   (delay (tc:hyp (:- pi (: exp exp-typ))))
				   (delay (tc:def (lambda (ts)
						    (tc:type-expr-bind
						     key-env-var
						     (tc:declaration-type-expr
						      (tc:env-update
						       '*tc:case-key* (tc:bind-type-expr 'lambda (car ts))
						       pi)))))))
				  (map (lambda (clause)
					 (delay (tc:hyp (:?- key-env-var (: clause (tc:basic-type-expr 'void))))))
				       clauses))
				 ;;----------------------------
				 (:- pi (: e *tc:void*)))))
	   (tc:and-exp
	    (position tests)
	    (tc:rule (map (lambda (test)
			    (:- pi (: test *tc:boolean*)))
			  tests)
		     ;;-------------------
		     (:- pi (: e *tc:boolean*))))
	   (tc:or-exp
	    (position tests)
	    (tc:rule (map (lambda (test)
			    (:- pi (: test *tc:boolean*)))
			  tests)
		     ;;-------------------
		     (:- pi (: e *tc:boolean*))))
	   (tc:let-exp
	    (position decls body)
	    (let ((binding-spec-vars (map tc:scheme-binding-spec->variable decls)))
	      (if (tc:no-duplicates? binding-spec-vars)
		  (let ((decl-types (tc:listof-new-variable-type-expr (length decls)))
			(env-var (tc:new-logicalvar))
			(body-type (tc:new-variable-type-expr)))
		    ;; delay by hand
		    (tc:rule-seq-local
		     (append
		      (map (lambda (decl type)
			     (delay (tc:hyp (:- pi (: decl type)))))
			   decls decl-types)
		      ;; Bind vars
		      (list
		       (delay (tc:def
			       (lambda (ts)
				 (tc:type-expr-bind
				  env-var
				  (tc:declaration-type-expr
				   (tc:env-join*
				    (append
				     (map tc:declaration-type-expr->env ts)
				     (list pi))))))))
		       (delay (tc:hyp
			       (:?- env-var (: body body-type))))))
		     ;;------------------------
		     (:- pi (: e body-type))))
		  ;; If there is a duplicate in the binding, we have an error
		  (tc:axiom (:- pi (: e (tc:duplicate-let-binding-error-maker
					 e (tc:find-duplicates binding-spec-vars))))))))
	   (tc:named-let-exp
	    (position name decls body)
	    ;; we desugar named let as in R5RS and check the
	    ;; resulting application of a letrec to the initial values
	    (let* ((formals (map tc:scheme-binding-spec->variable decls))
		   (exps (map tc:scheme-binding-spec->exp decls))
		   (lambda-exp
		    (tc:lambda-exp
		     position (tc:normal-formals position formals) body))
		   (letrec-exp (tc:letrec-exp
				position
				(list (tc:scheme-binding-spec
				       position name lambda-exp))
				(tc:scheme-body
				 position '()
				 (tc:procedure-call
				  position (tc:varref position name) exps)))))
	      (let ((type (tc:new-variable-type-expr)))
		(tc:rule (list
			  (:- pi (: letrec-exp type)))
			 ;;------------------------------
			 (:- pi (: e type))))))
	   (tc:let*-exp
	    (position decls body)
	    (let ((binding-spec-vars (map tc:scheme-binding-spec->variable decls)))
	      (if (tc:no-duplicates? binding-spec-vars)
		  (let* ((first-decl (car decls))
			 (first-decl-type (tc:new-variable-type-expr))
			 (other-decls (cdr decls))
			 (other-decl-types (tc:listof-new-variable-type-expr
					    (length other-decls)))
			 (env-vars (tc:new-lv-list (length other-decls)))
			 (last-env-var (tc:new-logicalvar))
			 (body-type (tc:new-variable-type-expr)))
		    (tc:rule-seq-local
		     (append
		      ;; Infer the first one
		      (list (delay (tc:hyp (:- pi (: first-decl first-decl-type)))))
		      ;;over the list
		      (apply append
			     (map (lambda (decl decl-type env-var)
				    (list 
				     (delay (tc:def
					     (lambda (ts)
					       (let ((definition-envs
						       (map tc:declaration-type-expr->env ts)))
						 (tc:type-expr-bind
						  env-var
						  (tc:declaration-type-expr
						   (tc:env-join*
						    (append
						     (map tc:declaration-type-expr->env ts)
						     (list pi)))))))))
				     (delay (tc:hyp (:?- env-var (: decl decl-type))))))
				  other-decls other-decl-types env-vars))	    
		      (list
		       ;; Get last var
		       (delay (tc:def
			       (lambda (ts)
				 (let ((definition-envs
					 (map tc:declaration-type-expr->env ts)))
				   (tc:type-expr-bind
				    last-env-var
				    (tc:declaration-type-expr
				     (tc:env-join*
				      (append
				       (map tc:declaration-type-expr->env ts)
				       (list pi)))))))))
		       ;; Infer body
		       (delay (tc:hyp
			       (:?- last-env-var (: body body-type))))))
		     ;;-----------------------------------------
		     (:- pi (: e body-type))))
		  ;; If there is a duplicate binding, we have an error
		  (tc:axiom (:- pi (: e (tc:duplicate-let-binding-error-maker
					 e (tc:find-duplicates binding-spec-vars))))))))
	   (tc:letrec-exp
	    (position decls body)
	    (let ((binding-spec-vars (map tc:scheme-binding-spec->variable decls)))
	      (if (tc:no-duplicates? binding-spec-vars)
		  (let* ((varrefs
			  (map (lambda (binding)
				 (tc:varref (tc:scheme-binding-spec->position binding)
					    (tc:scheme-binding-spec->variable binding)))
			       decls))
			 (vars (map tc:varref->variable varrefs))
			 (augmented-env (tc:env-join
					 (tc:env-update*
					  vars
					  (map (lambda (x)
						 (tc:bind-type-expr
						  'letrec
						  (tc:new-variable-type-expr)))
					       vars)
					  (tc:env-empty))
					 pi))
			 (decl-types (tc:listof-new-variable-type-expr (length decls)))
			 (env-var (tc:new-logicalvar))
			 (body-type (tc:new-variable-type-expr)))
		    ;; delay by hand
		    (tc:rule-seq-local
		     (append
		      (map (lambda (decl type)
			     (delay (tc:hyp (:- augmented-env (: decl type)))))
			   decls decl-types)
		      (list
		       (delay (tc:def
			       (lambda (ts)
				 (tc:type-expr-bind
				  env-var
				  (tc:declaration-type-expr
				   (tc:env-join*
				    (append
				     (map tc:declaration-type-expr->env ts)
				     (list pi))))))))
		       (delay (tc:hyp
			       (:?- env-var (: body body-type))))))
		     ;;------------------------
		     (:- pi (: e body-type))))
		  (tc:axiom (:- pi (: e (tc:duplicate-let-binding-error-maker e (tc:find-duplicates binding-spec-vars))))))))
	   (tc:begin-exp
	    (position commands exp)
	    (let ((last-type (tc:new-variable-type-expr)))
	      (tc:rule (append
			(map (lambda (command)
			       (:- pi (: command (tc:basic-type-expr 'void))))
			     commands)
			(list
			 (:- pi (: exp last-type))))
		       ;;--------------------------
		       (:- pi (: e last-type)))))

	   (tc:do-exp
	    (position iteration-specs test do-result commands)
	    (let ((iter-spec-vars
		   (map tc:scheme-iteration-spec->variable iteration-specs)))
	      (if (tc:no-duplicates? iter-spec-vars)
		  (let ((iter-var-types
			 (tc:listof-new-variable-type-expr
			  (length iteration-specs)))
			(iter-step-types
			 (tc:listof-new-variable-type-expr
			  (length iteration-specs)))
			(env-var (tc:new-logicalvar))
			(res-type (tc:new-variable-type-expr)))
		    (tc:rule-seq-local
		     (append
		      (map (lambda (iter-spec type)
			     (delay (tc:hyp (:- pi (: iter-spec type)))))
			   iteration-specs iter-var-types)
		      (list
		       (delay (tc:def
			       (lambda (ts)
				 (tc:type-expr-bind
				  env-var
				  (tc:declaration-type-expr
				  (tc:env-join*
				   (append
				    (map tc:declaration-type-expr->env ts)
				    (list pi)))))))) )
		      ;; the problem seems to in the next clause below;
		      ;; if you comment it out, everything else works fine
                      ;; (map (lambda (step type)
                      ;;      (delay (tc:hyp (:?- env-var (: step type)))))
                      ;;    (map tc:scheme-iteration-spec->step iteration-specs)
                      ;;    iter-step-types)
		      (map (lambda (iter-type step-type)
			     (delay (tc:hyp (<: step-type iter-type))))
			   iter-var-types iter-step-types)
		      (list
		       (delay (tc:hyp
			       (:?- env-var
				    (: test (tc:basic-type-expr 'boolean)))))
		       (delay (tc:hyp
			       (:?- env-var (: do-result res-type)))) )
		      (map (lambda (command)
			     (delay (tc:hyp
			       (:?- env-var
				    (: command (tc:basic-type-expr 'void))))))
			   commands) )
		     ;;--------------------------
		     (:- pi (: e res-type))))
		  (tc:axiom
		   (:- pi (: e (tc:duplicate-let-binding-error-maker
				e (tc:find-duplicates iter-spec-vars))))))))
	    
	   ;;	   (tc:do-stmt
	   ;;	    (position iteration-specs test commands) )

	   (tc:delay-exp
	    (position exp)
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: exp t)))
		       ;;-------------------------
		       (:- pi (: e (tc:applied-type-expr
				    (tc:basic-type-expr 'promise)
				    (list t)))))))
	   (tc:quasiquotation
	    (position datum)
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: datum t)))
		       ;;--------------------
		       (:- pi (: e t)))))
	   (tc:cases-exp
	    (position type-name exp clauses)
	    (tc:cases-exp-or-stmt-annotate
	     e pi type-name exp clauses))
	   (tc:cases-stmt
	    (position type-name exp clauses)
	    (tc:cases-exp-or-stmt-annotate
	     e pi type-name exp clauses))
	    
  	   ;; for the type checker
	   (tc:has-type-exp
	    (position type exp)
	    (let ((body-t (tc:new-variable-type-expr))
		  (decl-t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: exp body-t))
			(:- pi (: type decl-t))
			(<: body-t decl-t))
		       ;;------------------
		       (:- pi (: e decl-t)))))
	   (tc:has-type-trusted-exp
	    (position type exp)
	    (let ((t (tc:new-variable-type-expr)))
	      ;; Here we stop type checking and assume it's true
	      (tc:rule (list
			(:- pi (: type t)))
		       ;;--------------------
		       (:- pi (: e t)))))
	   (tc:test-type-exp
	    (position type-pred exp)
	    (let ((t (tc:new-variable-type-expr)))
	      ;; Here we just need a type predicate,
	      ;; which tests the exp at run-time
	      (tc:rule (list
			(:- pi (: type-pred
				  (tc:type-predicate-for-type-expr t))))
		       ;;------------------
		       (:- pi (: e t)))))
	   (tc:arrow-exp
	    (position arg-exps result-exp)
	    (let ((arg-types (tc:listof-new-variable-type-expr (length arg-exps)))
		  (result-type (tc:new-variable-type-expr)))
	      (tc:rule (append
			(map (lambda (arg type)
			       (:- pi (: arg (tc:type-predicate-for-type-expr type))))
			     arg-exps arg-types)
			(list
			 (:- pi (: result-exp (tc:type-predicate-for-type-expr result-type)))))
		       ;;-------------------------------------------
		       (:- pi (: e (tc:type-predicate-for-type-expr
				    (tc:function-type-expr arg-types result-type)))))))
	   (tc:varargs-arrow-exp
	    (position arg-exps repeated-exp result-exp)
	    (let ((arg-types (tc:listof-new-variable-type-expr (length arg-exps)))
		  (repeated-type (tc:new-variable-type-expr))
		  (result-type (tc:new-variable-type-expr)))
	      (tc:rule (append
			(map (lambda (arg type)
			       (:- pi (: arg (tc:type-predicate-for-type-expr type))))
			     arg-exps arg-types)
			(list
			 (:- pi (: repeated-exp (tc:type-predicate-for-type-expr repeated-type)))
			 (:- pi (: result-exp (tc:type-predicate-for-type-expr result-type)))))
		       ;;-------------------------------
		       (:- pi (: e (tc:type-predicate-for-type-expr
				    (tc:function-type-expr
				     (append arg-types
					     (list repeated-type (tc:basic-type-expr '...)))
				     result-type)))))))
	   (tc:forall-exp
	    (position formals exp)
	    (let ((formals-var (tc:new-variable-type-expr))
		  (exp-type (tc:type-predicate-for-type-expr (tc:new-variable-type-expr)))
		  (formals+pi-var (tc:new-logicalvar))
		  (formals-type-list '()))
	      (tc:rule-seq (list
			    (tc:hyp (:- pi (: formals formals-var)))
			    (tc:def
			     (lambda (ts)
			       (let* ((formals-env
				       (tc:env-map
					(lambda (k v)
					  (tc:bind-type-expr
					   'lambda
					   (tc:type-predicate-for-type-expr
					    (tc:binding->type-expr v))))
					(tc:declaration-type-expr->env (car ts))))
				      (formals+pi-env
				       (tc:env-join
					formals-env pi)))
				 (begin
				   (set! formals-type-list (tc:lookup-formals formals formals-env))
				   (tc:type-expr-bind formals+pi-var
						      (tc:declaration-type-expr formals+pi-env))))))
			    (tc:hyp (:?- formals+pi-var (: exp exp-type))))
			   ;;--------------------------------
			   (:- pi (: e (tc:function-type-expr
					formals-type-list
					exp-type))))))
	   ;; Used internally for the type checker
	   ;;[[[CHECK ME]]]
	   (tc:define-datatype-exp
	    (position variants)
	    (let ((ts (tc:listof-new-variable-type-expr (length variants))))
	      (tc:rule (map (lambda (variant t)
			      (:- pi (: variant t)))
			    variants ts)
		       ;;---------------------------
		       (:- pi (: e (tc:variant-record-type-expr ts))))))
	   (tc:constructor-for-variant-type-exp
	    (position argument-preds type-name)
	    (let* ((ts (tc:listof-new-variable-type-expr (length argument-preds)))
		   (pts (map tc:type-predicate-for-type-expr ts)))
	      (tc:rule (map (lambda (pred t)
			      (:- pi (: pred t)))
			    argument-preds pts)
		       ;;----------------------------
		       (:- pi (: e (tc:function-type-expr
				    ts
				    (tc:basic-type-expr type-name)))))))
	   (tc:module-exp
	    (position identifier initial-required-module-name
		      provides requires-for-syntax requires defreps
		      deftypes definitions)

	    (let* ((required-ts (tc:listof-new-variable-type-expr (length requires)))

		   ;; All of the concrete and abstract bindings for the module
		   (abs-deftypes  (tc:scheme-definitions->bindings deftypes #f '()))
		   (conc-deftypes (tc:scheme-definitions->bindings deftypes #f defreps))
		   (abs-deftype-ts (tc:listof-new-variable-type-expr (length abs-deftypes)))
		   (conc-deftype-ts (tc:listof-new-variable-type-expr (length conc-deftypes)))

		   ;; Environment variables for during the rule sequence
		   (reqs-env-var (tc:new-logicalvar))
		   (typedef-env-var (tc:new-logicalvar))
		   (result-env-var (tc:new-logicalvar))

		   ;; Bindings for the actual definitions
		   (bindings (tc:scheme-definitions->bindings definitions #f '() ))  
		   (binding-types (tc:listof-new-variable-type-expr (length bindings)))
		   (varrefs (map tc:scheme-binding-spec->varref bindings))
		   (vars (map tc:varref->variable varrefs))

		   ;; Build the recursive environment for bindings
		   (augmented-env (tc:env-update*
				   vars
				   (map (lambda (x)
					  (tc:bind-type-expr
					   'letrec
					   (tc:new-variable-type-expr)))
					vars)
				   (tc:env-empty))))
;;	       (displayln "REQS " requires)
	      (tc:rule-seq-local
	       (append
		;; Get all of the stuff from the requires
		(map (lambda (module-name type)
		       ; (displayln "MOD NAME" (tc:scheme-module-name->file-location module-name))
		       (delay
			 (tc:hyp
			  (:- (tc:env-empty)
			      (: (tc:import-from-module-name module-name)
				 type)))))
		     requires required-ts)
		;; Merge all of the required envs together
		(list (delay (tc:def
			      (lambda (ts)
				(let ((envs (map tc:declaration-type-expr->env ts)))
				  (tc:type-expr-bind
				   reqs-env-var
				   (tc:declaration-type-expr
				    (tc:env-join* envs))))))))
		;; Check for a definition for each provided module either locally or
		;; from the required modules.
		
		;; Infer the abstract deftypes
		(map (lambda (abs-deft type)
		       (delay (tc:hyp (:?- reqs-env-var (: abs-deft type)))))
		     abs-deftypes abs-deftype-ts)
		;; Now concrete bindings
		(map (lambda (conc-deft type)
		       (delay (tc:hyp (:?- reqs-env-var (: conc-deft type)))))
		     conc-deftypes conc-deftype-ts)		
		;;  Now create the environment in which we will do the checking
		;;  of the definitions
		(list (delay (tc:def
			      (lambda (ts)

;;				(displayln "T-ENVS")
;;				(map pretty-print (reverse ts))
				
				(let ((envs (cons augmented-env
						  (map tc:declaration-type-expr->env ts))))
				  (tc:type-expr-bind
				   typedef-env-var
				   (tc:declaration-type-expr
				    ;; the reverse here should deal with abstract bindings
				    (tc:env-join* (reverse envs))))))))) 
		;; Perform inference on the actual definitions now
		(map (lambda (def-binding type)
		       (delay (tc:hyp (:?- typedef-env-var (: def-binding type)))))
		     bindings binding-types)
		;; Build the final environment with all the results
		(list
		 (delay (tc:def
			 (lambda (ts)
			   (let ((envs (map tc:declaration-type-expr->env ts)))
			     (tc:type-expr-bind
			      result-env-var
			      (tc:declaration-type-expr
			       ;; Filter based on the provides
			       (tc:env-filter
				(lambda (k v)
				  (memq k provides))
				;; We return back to the abstract types by NOT reversing
				(tc:env-join* envs))))))))))
	       ;;------------------------------------------------------------
	       (:- pi (: e (tc:variable-type-expr result-env-var))))))

	   ;; Some unhandled expression
	   (else
	    (error
	     "tc:scheme-expression-annotate: cannot handle this expression: "
	     (tc:unparse-scheme-syntax e))))))

(define tc:infer-simple-datum-type
  (lambda (datum)
    ;; REQUIRES: datum is a scheme datum
    (cond
     ((number? datum) (tc:basic-type-expr 'number))
     ((char? datum) (tc:basic-type-expr 'char))
     ((string? datum) (tc:basic-type-expr 'string))
     ((boolean? datum) *tc:boolean*)
     ((symbol? datum) *tc:symbol*)
     ((null? (has-type datum datum))   
      (tc:applied-type-expr
       (tc:basic-type-expr 'list-of)
       (list (tc:new-variable-type-expr))))
     (else  ;; shouldn't happen 
      (error "tc:infer-simple-datum-type: unknown constant type for: "
	     (tc:unparse-scheme-syntax datum))))))

;; datums
(define tc:scheme-datum-annotate  
  (lambda (pi dat)
    (cases tc:scheme-datum dat
	   (tc:simple-datum
	    (position datum)
	    (let ((t (tc:infer-simple-datum-type datum)))
	      (tc:axiom (:- pi (: dat t)))))
	   (tc:normal-list-datum
	    (position elements)
	    (let ((t (tc:new-variable-type-expr))
		  (ts (tc:listof-new-variable-type-expr (length elements))))
	      (tc:rule-or
	       ;; no elements
	       (tc:rule-if '()               
			   (lambda (ts) (null? elements))
			   (lambda (ts) (tc:type-expr-null-subst))
			   ;; ------------------------------------
			   (:- pi (: dat (tc:make-applied-type-expr
					  (tc:make-basic-type-expr 'list-of)
					  (list (tc:new-variable-type-expr))))))
	       ;; all elements have the same type, t
	       (tc:rule-if (map
			    (lambda (element)
			      (:- pi (: element t)))
			    elements)
			   (lambda (ts)
			     (tc:all
			      (lambda (x)
				(equal? (car ts) x))
			      (cdr ts)))
			   (lambda (ts) (tc:type-expr-null-subst))
			   ;; ------------------------
			   (:- pi (: dat (tc:make-applied-type-expr
					  (tc:make-basic-type-expr 'list-of)
					  (list t)))))
	       ;; elements have different types
	       (tc:rule (map
			 (lambda (element type)
			   (:- pi (: element type)))
			 elements ts)
			;; -----------------------
			(:- pi (: dat (tc:make-applied-type-expr
				       (tc:make-basic-type-expr 'list-of)
				       (list *tc:datum*))))))))
	   (tc:improper-list-datum
	    (position before-dot after-dot)
	    ;; Where the before-dot list is always one item long (gauranteed by parser)
	    (let ((before-dot-elem (car before-dot))
		  (car-type (tc:new-variable-type-expr))
		  (cdr-type (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: before-dot-elem car-type))
			(:- pi (: after-dot cdr-type)))
		       ;;------------------------------------
		       (:- pi (: dat (tc:make-applied-type-expr
				      (tc:make-basic-type-expr 'pair-of)
				      (list car-type cdr-type)))))))
	   (tc:quoted-datum
	    (position datum)
	    (tc:rule-or
	     (tc:rule-if '()
			 (lambda (ts)
			   (and (tc:simple-datum? datum)
				(symbol? (tc:simple-datum->datum datum))))
			 (lambda (ts) (tc:type-expr-null-subst))
			 ;; ------------------------------------
			 (:- pi (: dat (tc:make-applied-type-expr
					(tc:make-basic-type-expr 'list-of)
					(list *tc:symbol*)))))
	     (tc:rule ()
		      ;; ----------------------------
		      (:- pi (: dat (tc:make-applied-type-expr
				     (tc:make-basic-type-expr 'list-of)
				     (list *tc:datum*)))))))
	   (tc:quasiquoted-datum
	    (position datum)
	    (if (and (tc:simple-datum? datum)
		     (symbol? (tc:simple-datum->datum datum)))
		(tc:axiom (:- pi (: dat (tc:make-applied-type-expr
					 (tc:make-basic-type-expr 'list-of)
					 (list *tc:symbol*)))))
		(tc:axiom (:- pi (: dat (tc:make-applied-type-expr
					 (tc:make-basic-type-expr 'list-of)
					 (list *tc:datum*)))))))
	   (tc:unquoted-datum
	    (position datum)
	    (if (and (tc:simple-datum? datum)
		     (symbol? (tc:simple-datum->datum datum)))
		(tc:axiom (:- pi (: dat (tc:make-applied-type-expr
					 (tc:make-basic-type-expr 'list-of)
					 (list *tc:symbol*)))))
		(tc:axiom (:- pi (: dat (tc:make-applied-type-expr
					 (tc:make-basic-type-expr 'list-of)
					 (list *tc:datum*)))))))
	   (tc:unquoted-exp
	    (position exp)
	    (let ((t1 (tc:new-variable-type-expr)))
	      ;; Pass on back to underlying expression
	      (tc:rule (list
			(:- pi (: exp t1)))
		       ;; -----------------------
		       (:- pi (: dat t1)))))
	   (tc:unquote-splicing-datum
	    (position datum)
	    (if (and (tc:simple-datum? datum)
		     (symbol? (tc:simple-datum->datum datum)))
		(tc:axiom (:- pi (: dat (tc:make-applied-type-expr
					 (tc:make-basic-type-expr 'list-of)
					 (list *tc:symbol*)))))
		(tc:axiom (:- pi (: dat (tc:make-applied-type-expr
					 (tc:make-basic-type-expr 'list-of)
					 (list *tc:datum*)))))))
	   (tc:unquote-splicing-exp
	    (position exp)
	    (let ((t1 (tc:new-variable-type-expr))
		  (t2 (tc:make-applied-type-expr
		       (tc:make-basic-type-expr 'list-of)
		       (list (tc:new-variable-type-expr)))))
	      (tc:rule (list
			(:- pi (: exp t1))
			(<: t1 t2))
		       ;; -------------------
		       (:- pi (: dat t1)))))
	   (tc:vector-datum
	    (position elements)
	    (let ((t (tc:new-variable-type-expr))
		  (ts (tc:listof-new-variable-type-expr (length elements))))
	      (tc:rule-or
	       ;; no elements
	       (tc:rule-if '()    ;; no hypotheses needed
			   (lambda (ts) (null? elements))
			   (lambda (ts) (tc:type-expr-null-subst))
			   ;; ------------------------------------
			   (:- pi (: dat (tc:make-applied-type-expr
					  (tc:make-basic-type-expr 'vector-of)
					  (list (tc:new-variable-type-expr)  )))))
	       ;; all elements have the same type, t
	       (tc:rule-if (map
			    (lambda (element)
			      (:- pi (: element t)))
			    elements)
			   (lambda (ts)
			     (tc:all
			      (lambda (x)
				(equal? (car ts) x))
			      (cdr ts)))
			   (lambda (ts) (tc:type-expr-null-subst))
			   ;; ------------------------
			   (:- pi (: dat (tc:make-applied-type-expr
					  (tc:make-basic-type-expr 'vector-of)
					  (list t)))))
	       ;; elements have different types
	       (tc:rule (map
			 (lambda (element type)
			   (:- pi (: element type)))
			 elements ts)
			;; -----------------------
			(:- pi (: dat (tc:make-applied-type-expr
				       (tc:make-basic-type-expr 'vector-of)
				       (list *tc:datum*))))))))
    (else (error "tc:scheme-datum-annotate: Cannot handle datum: "
		 (tc:unparse-scheme-syntax dat))))))

;;formals
(define tc:scheme-formals-annotate
  (lambda (pi e)
    (cases tc:scheme-formals e
	   (tc:normal-formals
	    (position variables)
	    (let ((variable-types (tc:listof-new-variable-type-expr
				   (length variables))))
	      (tc:rule-if
	       '() ;; No Hypotheses
	       (lambda (ts)
		 (tc:no-duplicates? variables))
	       (lambda (ts) (tc:type-expr-null-subst))
	       ;;---------------------
	       (:- pi (: e (tc:declaration-type-expr
			    (tc:env-update*
			     variables
			     (map (lambda (v) (tc:bind-type-expr 'lambda v))
				  variable-types)
			     (tc:env-empty))))))))	    
	   (tc:unrestricted-formals
	    (position variable)
	    (let* ((elem-type (tc:new-variable-type-expr))
		   (impl-formal-type (tc:applied-type-expr
				      (tc:basic-type-expr 'list-of)
				      (list elem-type))))
	      (tc:axiom (:- pi (: e (tc:declaration-type-expr
				     (tc:env-update
				      variable
				      (tc:bind-type-expr 'lambda impl-formal-type)
				      (tc:env-empty))))))))			     
	   (tc:some-normal-formals
	    (position before-dot after-dot)
	    (let ((before-dot-types (tc:listof-new-variable-type-expr
				     (length before-dot)))
		  (after-dot-type (tc:applied-type-expr
				   (tc:basic-type-expr 'list-of)
				   (list (tc:new-variable-type-expr)))))
	      (tc:rule-if
	       '() ;; No Hypotheses
	       (lambda (ts)
		 (tc:no-duplicates? (append before-dot (list after-dot))))
	       (lambda (ts) (tc:type-expr-null-subst))
	       ;;-----------------------
	       (:- pi (: e (tc:declaration-type-expr
			    (tc:env-update*
			     before-dot
			     (map (lambda (v) (tc:bind-type-expr 'lambda v))
				  before-dot-types)
			     (tc:env-update
			      after-dot
			      (tc:bind-type-expr 'lambda after-dot-type)
			      (tc:env-empty)))))))))
	   (else
	    (error
	     "tc:scheme-formals-annotate: cannot annotate expr: "
	     (tc:unparse-scheme-syntax e))))))

;;bodies
(define tc:scheme-body-record-annotate
  (lambda (pi e)
    (cases tc:scheme-body-record e
	   (tc:scheme-body
	    (position definitions sequence)
	    (let ((t (tc:new-variable-type-expr))
		  (dt (tc:new-variable-type-expr))
		  (defs-env-var (tc:new-logicalvar)))
	      (tc:rule-seq-local
	       ;; Must delay manually since we don't match syntax macro
	       (append
		(map (lambda (definition)                             ;;LIKE LETREC
		       (delay (tc:hyp (:- pi (: definition dt)))))
		     definitions)
		(list ;; [[[FIXME: to make this more sound, should check for duplicate definitions]]]
		 (delay (tc:def (lambda (ts)
				  (let ((definition-envs (map tc:declaration-type-expr->env ts)))
				    (tc:type-expr-bind
				     defs-env-var
				     (tc:declaration-type-expr
				      (tc:env-join*
				       (reverse (cons pi definition-envs))))))))) ;; Pi has least precedence, followed by others
		 (delay (tc:hyp (:?- defs-env-var (: sequence t))))))
	       ;;-------------------
	       (:- pi (: e t))))))))
  
;;[[[FIXME: can we do this with rule-ors?  Had problem with seq]]]
;;definitions
(define tc:scheme-definition-annotate
  (lambda (pi e)
    (cases tc:scheme-definition e
	   (tc:simple-define
	    (position variable exp)
	    (let ((t (tc:new-variable-type-expr))
		  (pi+var (tc:env-update
			   variable
			   (tc:bind-type-expr
			    'letrec
			    (tc:new-variable-type-expr))
			   pi)))
	      (if (and (tc:env-bound? variable pi)
		       (not (eq? 'letrec (tc:binding->binder (tc:env-value variable pi)))))
		  ;; If the variable is bound already, use it
		  (tc:rule (list
			    (:- pi (: exp t))
			    (<: t (tc:binding->type-expr (tc:env-value variable pi))))
			   ;;------------------
			   (:- pi (: e (tc:declaration-type-expr
					(tc:env-empty)))))
		  ;; Otherwise use inferred type
		  (tc:rule (list
			    (:- pi+var (: exp t)))
			   ;;---------------------
			   (:- pi (: e (tc:declaration-type-expr
					(tc:env-update
					 variable
					 (tc:bind-type-expr 'let t)
					 (tc:env-empty)))))))))
	   (tc:procedure-define
	    (position variable formals body)
	    (let ((named-lambda-exp (tc:named-lambda-exp position variable formals body))
		  (lambda-exp-type (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: named-lambda-exp lambda-exp-type)))
		       ;;-------------------------------------------
		       (:- pi (: e (tc:declaration-type-expr
				    (tc:env-update
					     variable
					     (tc:bind-type-expr 'let lambda-exp-type)
					     (tc:env-empty))))))))

	   ;; [[[CHECK ME--how do I make this happen--think this is wrong?]]]
	   (tc:grouped-define
	    (position definitions)
	    (let ((definitions-var (tc:new-variable-type-expr))
		  (merged-envs-var (tc:new-logicalvar)))
	      (tc:rule-seq-local ;; Since we don't match the syntax pattern for rule-seq, do delay manually
	       (append                     
		(map (lambda (definition)
		       (delay (tc:hyp (:- pi (: definition definitions-var)))))
		     definitions)
		(list (delay (tc:def
			      (lambda (ts)
				(let ((definition-envs (map tc:declaration-type-expr->env ts)))
				  (tc:type-expr-bind				     
				   merged-envs-var
				   (tc:declaration-type-expr
				    (tc:env-join* (reverse definition-envs))))))))))
	       ;; Last definition has precedence
	       ;;--------------------------------------
	       (:?- pi (: e (tc:declaration-type-expr merged-envs-var))))))
	   
	   (tc:define-datatype
	    (position type-name type-predicate-name variants)
	    (let ((ts (tc:listof-new-variable-type-expr (length variants)))
		  (augmented-env (tc:env-update
				  type-predicate-name
				  (tc:bind-type-expr
				   'let
				   (tc:type-predicate-for-type-expr
				    (tc:basic-type-expr type-name)))
				  pi))
		  (variant-types '()))
	      ;; If already bound (other than letrec), error
	      (tc:rule-if (map (lambda (variant t)
				 (:- augmented-env (: variant t)))
			       variants ts)
			  (lambda (ts)
			    (or (not (tc:env-bound? type-name pi))
				(eq? 'letrec (tc:binding->binder (tc:env-value type-name pi)))))
			  (lambda (ts)
			    (begin			  
			      (set! variant-types ts)
			      (tc:type-expr-null-subst)))
			  ;;------------------------------------
			  (:- pi (: e (tc:declaration-type-expr
				       (tc:env-update*
					(append
					 (list type-name)
					 (list type-predicate-name)
					 (map tc:variant-decl->variant-name variants))
					(append
					 (list (tc:bind-type-expr 'let
								  (tc:variant-record-type-expr ts)))
					 (list (tc:bind-type-expr 'let
								  (tc:type-predicate-for-type-expr
								   (tc:basic-type-expr type-name))))
					 (map (lambda (variant)
						(tc:bind-type-expr
						 'let
						 (tc:function-type-expr
						  (map tc:field-type-binding-type-expr->type
						       (tc:variant-type-expr->fields variant))
						  (tc:basic-type-expr type-name))))
					      variant-types))
					(tc:env-empty))))))))
	   (tc:deftype
	    (position variable type)
	    (let ((t (tc:new-variable-type-expr)))
	      ;; If the variable is already bound (non letrec), we have an error 
	      (tc:rule-if
	       (list (:- pi (: type t)))
	       (lambda (ts)
		 (or (not (tc:env-bound? variable pi))
		     (eq? 'letrec (tc:binding->binder (tc:env-value variable pi)))))
	       (lambda (ts) (tc:type-expr-null-subst))
	       ;;-----------------------------
	       (:- pi (: e (tc:declaration-type-expr
			    (tc:env-update
			     variable
			     (tc:bind-type-expr 'let t)
			     (tc:env-empty))))))))  	    
	   (else (error "tc:scheme-definition-annotate cannot handle: "
			(tc:unparse-scheme-syntax e))))))

(define tc:variant-decl-record-annotate
  (lambda (pi e)
    (cases tc:variant-decl-record e
	   (tc:variant-decl
	    (position variant-name fields)
	    (let ((ts (tc:listof-new-variable-type-expr (length fields))))
	      (tc:rule (map (lambda (field t)
			      (:- pi (: field t)))
			    fields ts)
		       ;;-----------------------------
		       (:- pi (: e (tc:variant-type-expr variant-name ts)))))))))

(define tc:field-pred-binding-record-annotate
  (lambda (pi e)
    (cases tc:field-pred-binding-record e
	   (tc:field-pred-binding
	    (position field-name predicate-exp)
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: predicate-exp (tc:type-predicate-for-type-expr t))))  
		       ;;--------------------------
		       (:- pi (: e (tc:field-type-binding-type-expr
				    field-name t)))))))))

(define tc:scheme-cond-clause-annotate
  (lambda (pi e)
    (cases tc:scheme-cond-clause e
	   (tc:normal-cond-clause
	    (position test sequence)
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: test *tc:boolean*))
			(:- pi (: sequence t)))
		       ;;----------------------
		       (:- pi (: e t)))))
	   (tc:test-only-cond-clause
	    (position test)
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: test t)))
		       ;;-----------------
		       (:- pi (: e t)))))
	   (tc:test=>cond-clause
	    (position test recipient)
	    (let ((recp-type (tc:new-variable-type-expr))
		  (result-type (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: test *tc:boolean*))
			(:- pi (: recipient recp-type))
			(<: recp-type (tc:function-type-expr
				       (list *tc:boolean*)
				       result-type)))
		       ;;-----------------------
		       (:- pi (: e result-type)))))
	   (tc:else-cond-clause
	    (position sequence)
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: sequence t)))
		       ;;----------------------
		       (:- pi (: e t)))))
	   (else (error "tc:scheme-cond-clause-annotate cannot handle: "
			(tc:unparse-scheme-syntax e))))))

(define tc:scheme-case-clause-annotate
  (lambda (pi e)
    (cases tc:scheme-case-clause e
	   (tc:normal-case-clause
	    (position datums sequence)
	    (let ((key-type (tc:binding->type-expr (tc:env-value '*tc:case-key* pi)))
		  (seq-type (tc:new-variable-type-expr)))
	      (tc:rule (append
			(map (lambda (datum)			   
			       (:- pi (: datum key-type)))
			     datums)
			(list
			 (:- pi (: sequence seq-type))))
		       ;;--------------------------
		       (:- pi (: e seq-type)))))
	   (tc:else-case-clause
	    (position sequence)
	    (let ((seq-type (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: sequence seq-type)))
		       ;;------------------------
		       (:- pi (: e seq-type))))))))

(define tc:scheme-binding-spec-annotate
  (lambda (pi e)
    (cases tc:scheme-binding-spec-record e
	   (tc:scheme-binding-spec
	    (position variable exp)
	    (if (and (tc:env-bound? variable pi)
		     (not (eq? 'letrec (tc:binding->binder (tc:env-value variable pi)))))
		(let ((t (tc:binding->type-expr (tc:env-value variable pi))))
		  (tc:rule (list
			    (:- pi (: exp t)))
			   ;;--------------------------
			   (:- pi (: e (tc:declaration-type-expr
					(tc:env-update
					 variable
					 (tc:bind-type-expr 'let t)
					 (tc:env-empty)))))))
		;; Not bound yet
		(let ((t (tc:new-variable-type-expr)))		  
		  (tc:rule (list
			    (:- pi (: exp t)))
			   ;;------------------------
			   (:- pi (: e (tc:declaration-type-expr
					(tc:env-update
					 variable
					 (tc:bind-type-expr 'let t)
					 (tc:env-empty))))))))))))

(deftype tc:scheme-cases-clause-annotate
  (-> (tc:type-env tc:scheme-cases-clause)
      (tc:atree-of tc:scheme-cases-clause)))
(define tc:scheme-cases-clause-annotate
  (lambda (pi e)
    (cases tc:scheme-cases-clause e
	   (tc:normal-cases-clause
	    (position variant-name fields sequence)
	    (let ((field-types ;; pull field types out of environment
 		   (tc:function-type-expr->arg-types 
		    (tc:binding->type-expr (tc:env-value variant-name pi)))))
	      (if (and (tc:no-duplicates? fields)
		       (= (length fields) (length field-types)))
		  (let ((t (tc:new-variable-type-expr))
			(augmented-env
			 (tc:env-update*
			  fields
			  (map (lambda (type)
				 (tc:bind-type-expr 'lambda type))
			       field-types)
			  pi)))
		    (tc:rule (list
			      (:- augmented-env (: sequence t)))
			     ;;---------------------
			     (:- pi (: e t))))
		  ;; mismatch in number of fields or duplicates in parameters
		  (tc:axiom (:- pi (: e (tc:cases-clause-bad-fields-error-maker e fields field-types)))))))
	   (tc:else-cases-clause
	    (position sequence)
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: sequence t)))
		       ;;---------------------
		       (:- pi (: e t))))))))

(define tc:cases-exp-or-stmt-annotate
  (lambda (e pi type-name exp clauses)
;    (let* ((varref-exp (tc:varref (tc:scheme-syntax->position e) type-name))
;	   (clause-var-types (tc:listof-new-variable-type-expr (length clauses)))
;	   (generate-result (tc:generate-subtype-rules clause-var-types))
;	   (subtype-rules (map tc:hyp (car generate-result)))
;	   (supertype (tc:variable-type-expr (cddr generate-result))))
      
;      (tc:rule-seq-local
;       (append
;	(list (delay (tc:hyp (:- pi (: varref-exp (tc:new-variable-type-expr)))))
;	      (delay (tc:when
;		      (lambda (ts)
;			(displayln "TS " ts)
;			(displayln "RES " (and (tc:variant-record-type-expr? (car ts))
;			     (tc:check-cases-clause-usage (car ts) clauses)))		
;			(and (tc:variant-record-type-expr? (car ts))
;			     (tc:check-cases-clause-usage (car ts) clauses)))))
;	      (delay (tc:hyp (:- pi (: exp (tc:basic-type-expr type-name))))))
;	(map (lambda (clause t)
;	       (delay (tc:hyp (:- pi (: clause t)))))
;	     clauses clause-var-types)
;	subtype-rules)
;       ;;-----------------------------
;       (:- pi (: e supertype))))))
		 	     
    (if (and (tc:env-bound? type-name pi)
	     (tc:variant-record-type-expr? (tc:binding->type-expr (tc:env-value type-name pi))))
	(let ((variant-type (tc:binding->type-expr (tc:env-value type-name pi))))
	  ;; Verify that a case appears for each variant and that there
	  ;; are no duplicates
	  (if (tc:check-cases-clause-usage variant-type clauses)  
	      (let* ((clause-var-types (tc:listof-new-variable-type-expr (length clauses)))
		     (generate-result (tc:generate-subtype-rules clause-var-types))
		     (subtype-rules (car generate-result))
		     (supertype (tc:variable-type-expr (cddr generate-result))))
		(tc:rule (append
			  (list (:- pi (: exp (tc:basic-type-expr type-name))))
			  (map (lambda (clause t)
				 (:- pi (: clause t)))
			       clauses clause-var-types)
			  subtype-rules)
		 ;;------------------------------
			 (:- pi (: e supertype))))
	      ;; bad usage of clauses
	      (tc:axiom (:- pi (: e (tc:cases-with-bad-clause-usage-error-maker
				     e variant-type))))))
	;; can't cases on non-variant
	(tc:axiom (:- pi (: e (tc:cases-on-non-variant-record-error-maker
			       e type-name)))))))
	   
(define tc:scheme-iteration-spec-annotate
  (lambda (pi e)
    (cases tc:scheme-iteration-spec e
	   (tc:stepped-iteration-spec
	    (position variable init step)
	    ;; step is ignored here, becuase
	    ;; pi isn't the right environment to check it in
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: init t)))
		       ;;---------------------
	       (:- pi (: e (tc:declaration-type-expr
			    (tc:env-update
			     variable
			     (tc:bind-type-expr 'lambda t)
			     (tc:env-empty))))))))
	   (tc:init-only-iteration-spec
	    (position variable init)
	    (let ((t (tc:new-variable-type-expr)))
	      (tc:rule (list
			(:- pi (: init t)))
		       ;;---------------------
		       (:- pi (: e (tc:declaration-type-expr
				    (tc:env-update
				     variable
				     (tc:bind-type-expr 'lambda t)
				     (tc:env-empty)))))))))))

(define tc:input-type-exp-annotate
  (lambda (pi e)
    (cases tc:input-type-exp e
	   (tc:translated-input-type-exp	    
	    (position type-expr)
	    (tc:axiom (:- pi (: e type-expr))))
	   (else
	    (tc:axiom (:- pi (: e (tc:input-type-parse e))))))))


;;; ------------- HELPERS -----------------
(deftype tc:subst-bindings-c 
  (-> (tc:type-env)
      (-> (tc:type-expr) tc:type-expr)))
(define tc:subst-bindings-c
  (lambda (env)
    (letrec ((subst-bindings
	      (lambda (x)
		;; EFFECT: returns a tc:type-expr like x but with each
		;; variable bound in env replaced by its binding and
		;; recursively bound variables accounted for
		(cond
		 ((and (tc:variable-type-expr? x)
		       (tc:env-bound? (tc:variable-type-expr->uniq-sym x)
				      env))
		  (subst-bindings
		   (tc:env-value (tc:variable-type-expr->uniq-sym x) env)))
		 ((or (tc:variable-type-expr? x) (tc:basic-type-expr? x)) x)
		 ((tc:function-type-expr? x)
		  (tc:make-function-type-expr
		   (map subst-bindings
			(tc:function-type-expr->arg-types x))
		   (subst-bindings (tc:function-type-expr->result-type x))))
		 ((tc:intersection-type-expr? x)
		  (tc:make-intersection-type-expr
		   (map subst-bindings
			(tc:intersection-type-expr->conjoined-types x))))
		 ((tc:type-predicate-for-type-expr? x)
		  (tc:make-type-predicate-for-type-expr
		   (subst-bindings (tc:type-predicate-for-type-expr->type x))))
		 ((tc:variant-record-type-expr? x)
		  (tc:make-variant-record-type-expr
		   (map subst-bindings
			(tc:variant-record-type-expr->variants x))))
		 ((tc:variant-type-expr? x)
		  (tc:make-variant-type-expr
		   (tc:variant-type-expr->variant-name x)
		   (map subst-bindings
			(tc:variant-type-expr->fields x))))
		 ((tc:field-type-binding-type-expr? x)
		  (tc:make-field-type-binding-type-expr
		   (tc:field-type-binding-type-expr->field-name x)
		   (subst-bindings (tc:field-type-binding-type-expr->type x))))
		 ((tc:applied-type-expr? x)
		  (tc:make-applied-type-expr
		   (tc:applied-type-expr->operator-type x)
		   (map subst-bindings
			(tc:applied-type-expr->operand-types x))))

		 ;; [[[FIXME, what to do here?]]]
		 ((tc:declaration-type-expr? x)
		  (tc:declaration-type-expr   ;;;HUH???		   
		   (tc:env-map subst-bindings (tc:declaration-type-expr->env x))))
		 
		 ((tc:error-type-expr? x) x)
		 (else
		  (error "tc:subst-bindings-c, illegal type syntax: " x))))))
      subst-bindings)))
      
(deftype tc:subst-bindings
  (-> (tc:type-env tc:type-expr) tc:type-expr))
(define tc:subst-bindings
  (lambda (bindings x)
    ;; EFFECT: returns a tc:type-expr like x but with each tc:variable 
    ;;         bound in env replaced by its binding and 
    ;;         recursively bound tc:variables accounted for
    ;; (displayln "tc:subst-bindings with bindings " bindings " and x = " x)
    ((tc:subst-bindings-c bindings) x)))
					     
(define tc:lookup-formals
  (lambda (formals formals-env)
    (letrec ((list-lookup
	      (lambda (vars accum-list)
		(if (null? vars)
		    accum-list
		    (let ((type (tc:binding->type-expr (tc:env-value (car vars) formals-env))))
		      (list-lookup
		       (cdr vars)
		       (append accum-list (list type))))))))		     
      (cases tc:scheme-formals formals
	   (tc:normal-formals
	    (position variables)
	    (list-lookup variables '()))	   
	   (tc:unrestricted-formals
	    (position variable)
	    (let ((elem-type
		   (car (tc:applied-type-expr->operand-types
			 (car (list-lookup (list variable) '()))))))
	      (list elem-type (tc:make-basic-type-expr '...))))	      
	   (tc:some-normal-formals
	    (position before-dot after-dot)
	    (let ((before-dot-types (list-lookup before-dot '()))
		  (after-dot-elem-type
		   (car (tc:applied-type-expr->operand-types
			 (car (list-lookup (list after-dot) '()))))))
	      (append before-dot-types
		      (list after-dot-elem-type
			    (tc:make-basic-type-expr '...)))))))))

(define tc:generate-subtype-rules
  (letrec ((iter
	    (lambda (types rules-list last-subtypevar last-supertype-var)
	      (if (< (length types) 2)
		  (cons rules-list (cons last-subtypevar last-supertype-var))
		  (let ((subtype-var (tc:new-logicalvar))
			(supertype-var (tc:new-logicalvar)))
		    (iter
		     (cons (tc:variable-type-expr supertype-var) (cddr types))
		     (append rules-list
			     (list (<:> (car types) (cadr types)
					subtype-var supertype-var)))
		     subtype-var
		     supertype-var))))))
    (lambda (types)
      ;;REQUIRES: types has at least 1 element
      ;;ENSURES: Returns is a pair whose first element is a list of
      ;;         subtyping rules created by taking elements pairwise from
      ;;         types.  The second element is itself a pair contianing
      ;;         the variables which were bound to the bottommost type and
      ;;         topmost type in the list. If types does not contain at least
      ;;         2 elements, the resulting rules bind the sub and super type
      ;;         variables to the only type in the list.
      ;;         The rules created only specify that one is a
      ;;         subtype of the other, but not direction.
      (let ((sub-var (tc:new-logicalvar))
	    (sup-var (tc:new-logicalvar)))
	(if (< (length types) 2)
	    (cons
	     (list
	      (<: (tc:variable-type-expr sub-var) (car types))
	      (<: (tc:variable-type-expr sup-var) (car types)))
	     (cons sub-var sup-var))
	    (iter types '()
		  ;; These will get thrown away, but for type correctness
		  sub-var sup-var))))))

;; no duplicates
;; names match variant-type
(define tc:check-cases-clause-usage
  (lambda (variant-type clauses)
    (letrec ((else-clause-occurs
	      (not (null? (tc:filter tc:else-cases-clause? clauses))))
	     (get-clause-names
	      (lambda (clauses accum-list)
		(cond
		 ((null? clauses) accum-list)
		 ;;else clause must occur last from parser, so ignore rest
		 ((tc:else-cases-clause? (car clauses)) accum-list)
		 (else
		  (cases tc:scheme-cases-clause (car clauses)
			 (tc:normal-cases-clause
			  (position variant-name fields sequence)
			  (get-clause-names (cdr clauses) (append accum-list (list variant-name))))
			 (else (error ("tc:check-cases-clause-usage invalid clause type: " (car clauses))))))))))
      (let ((clause-names (get-clause-names clauses '()))
	    (variant-decl-names (map tc:variant-type-expr->variant-name
				     (tc:variant-record-type-expr->variants variant-type))))
	(and (tc:no-duplicates? clause-names)
	     ;; We have all of the variant cases
	     (or (tc:set-equal? clause-names variant-decl-names)
		 ;; We have some of the cases and an else
		 (and (tc:set-subset? clause-names variant-decl-names)
		      else-clause-occurs)))))))

(define tc:scheme-forms->requires
  (lambda (forms)
    (apply append
	   (map tc:scheme-form->requires forms))))

(define tc:scheme-form->requires
  (lambda (form)
    (cases tc:scheme-form form
	   (tc:load-command (position operator filename program)
			    (tc:scheme-program->requires program))
	   (else '()))))

(define tc:scheme-program->requires
  (lambda (prog)
    (cases tc:scheme-program prog
	   (tc:scheme-program-normal
	    (position require-forms deftypes forms)
	    (append require-forms (tc:scheme-forms->requires forms))))))

(define tc:scheme-forms->deftypes
  (lambda (forms)
    (apply append
	   (map tc:scheme-form->deftypes forms))))

(define tc:scheme-form->deftypes
  (lambda (form)
    (cases tc:scheme-form form
	   (tc:load-command (position operator filename program)
			    (tc:scheme-program->deftypes program))
	   (else '()))))

(define tc:scheme-program->deftypes
  (lambda (prog)
    (cases tc:scheme-program prog
	   (tc:scheme-program-normal
	    (position require-forms deftypes forms)
	    (append deftypes (tc:scheme-forms->deftypes forms))))))

(define tc:scheme-forms->commands
  (lambda (forms)
    (apply append
	   (map tc:scheme-form->commands forms))))

(define tc:scheme-form->commands
  (lambda (form)
    (cases tc:scheme-form form
	   (tc:command 
	    (position exp) (list form))
	   (tc:load-command (position operator filename program)
			    (tc:scheme-program->commands program))
	   (else '()))))

(define tc:scheme-program->commands
  (lambda (prog)
    (cases tc:scheme-program prog
	   (tc:scheme-program-normal
	    (position require-forms deftypes forms)
	    (tc:scheme-forms->commands forms)))))

(define tc:scheme-program->bindings
  (lambda (prog)
    (cases tc:scheme-program prog
	   (tc:scheme-program-normal
	    (position require-forms deftypes forms)
	    (tc:scheme-forms->bindings forms #f '())))))

(deftype tc:scheme-form->bindings
 (-> (tc:scheme-form
      boolean
      (list-of (pair-of tc:input-type-exp tc:input-type-exp)))
     (list-of tc:scheme-binding-spec)))
(define tc:scheme-form->bindings
  (lambda (form trusted defreps)
    ;; REQUIRES: if trusted, then defreps = ()
    ;; ENSURES: if not trusted, returns a list of bindings
    ;;          corresponding to top-level definitions, downed by the defreps,
    ;;          otherwise returns a list of all trusted bindings
    (cases tc:scheme-form form
      (tc:definition (position definition)
	(tc:scheme-definition->bindings
	 definition trusted defreps))
      (tc:load-command (position operator filename program)
		       (tc:scheme-program->bindings program))
;;	(if trusted
;;	    (tc:scheme-program->bindings program)
;;	    '()))
      (tc:module
       (position identifier initial-required-module-name
		 provides requires-for-syntax requires defreps
		 deftypes definitions)
       (list 
       (tc:scheme-binding-spec
	position identifier
	(tc:module-exp position identifier initial-required-module-name
		       provides requires-for-syntax requires defreps
		       deftypes definitions))))
      ;; [[FIXME]]
;;      (tc:begin-form (position forms)
;;	(if trusted
;;	    (let ((publics (tc:scheme-forms->publics forms)))
;;	      (maybe-test
;;	       publics
;;	       (lambda (public-ids)
;;		 (tc:filter
;;		  (lambda (bs)
;;		    (memq (tc:scheme-binding-spec->variable bs) public-ids))
;;		  (tc:scheme-forms->bindings
;;		   forms trusted defreps)))
;;	       (lambda () (tc:scheme-forms->bindings
;;			   forms trusted defreps))))
;;	    '()))
      (else '()))))

(deftype tc:scheme-forms->bindings
 (-> ((list-of tc:scheme-form)
      boolean
      (list-of (pair-of tc:input-type-exp tc:input-type-exp)))
     (list-of tc:scheme-exp)))
(define tc:scheme-forms->bindings
  (lambda (forms trusted defreps)
    ;; ENSURES: if not trusted, returns a list of bindings
    ;;          corresponding to top-level definitions, downed by the defreps,
    ;;          otherwise returns a list of all trusted bindings
    (apply append
	   (map (lambda (f)(tc:scheme-form->bindings
			    f trusted defreps))
		forms))))

(deftype tc:scheme-definition->bindings
 (-> (tc:scheme-definition
      boolean
      (list-of (pair-of tc:input-type-exp tc:input-type-exp)))
     (list-of tc:scheme-binding-spec)))
(define tc:scheme-definition->bindings
  (lambda (definition trusted defreps)
    ;; ENSURES: returns a list bindings corresponding to top-level definitions
    ;;          but only the trusted ones if trusted is true,
    ;;          and downed by defreps
    ;; Notes: The defreps are used in the generation of the expressions
    ;;        corresponding to the deftypes and define-record-types.
    ;;        Hence, if defreps is null, then then the result is "abstract",
    ;;        otherwise it is "concrete".
    (cases tc:scheme-definition definition
      (tc:simple-define (position variable exp)
	(if trusted
	    '()
	    (list (tc:scheme-binding-spec position variable exp))))
      (tc:procedure-define (position variable formals body)
	(if trusted
	    '()
	    (list (tc:scheme-binding-spec
		   position
		   variable
		   (tc:named-lambda-exp position variable formals body)))))
      (tc:deftype (position variable type)
	;; used both if trusted and not
	(list (tc:make-binding-for-deftype position variable type defreps)))
      (tc:define-datatype (position type-name type-predicate-name variants)
	(if trusted
	    '()
	    (cons
	     ;; binding for the type-predicate-name procedure
	     (tc:make-binding-for-deftype
	      position type-predicate-name
	      (tc:type-predicate-for-input-type-exp
	       position (tc:basic-input-type-exp position type-name))
	      '())
	     (cons
	      ;; binding for the type-name
	      (tc:scheme-binding-spec
	       position type-name
	       (tc:define-datatype-exp position variants))
	      ;; bindings for the variant constructors
	      (map (tc:variant-decl->constructor-binding-maker type-name)
		   variants)))))
      (tc:grouped-define (position definitions)
        ;; this isn't used, because the parser eliminates grouped-defines
	(error "tc:scheme-definition->bindings saw a grouped-define!"
	       definition))
      (else (error "tc:scheme-definition->bindings" definition)))))

(deftype tc:scheme-definitions->bindings
 (-> ((list-of tc:scheme-definition)
      boolean
      (list-of (pair-of tc:input-type-exp tc:input-type-exp)))
     (list-of tc:scheme-exp)))
(define tc:scheme-definitions->bindings
  (lambda (definitions trusted defreps)
    (apply append
	   (map (lambda (d)
		  (tc:scheme-definition->bindings
		   d trusted defreps))
		definitions))))

(deftype tc:make-binding-for-deftype
  (-> (tc:position
       symbol
       tc:input-type-exp
       (list-of (pair-of tc:input-type-exp tc:input-type-exp)))
      tc:scheme-binding-spec))
(define tc:make-binding-for-deftype
  (let ((string-val "tc:make-binding-for-deftype"))
    (lambda (position variable type defreps)
      (tc:scheme-binding-spec
       position
       variable
       (tc:has-type-trusted-exp
	position
	(tc:mult-def-down-type defreps type)
	(tc:self-evaluating position string-val))))))

(define tc:variant-decl->constructor-binding-maker
  (lambda (type-name)
    (lambda (vd)
      (cases tc:variant-decl-record vd
       (tc:variant-decl (position variant-name fields)
   	 (tc:scheme-binding-spec
	  position variant-name
	  (tc:constructor-for-variant-type-exp
	   position
	   (map tc:field-pred-binding->predicate-exp fields)
	   type-name)))))))

;;;--------------Down Type Helpers for Defrep--------------------------

(define tc:infer-input-type-exp
  (lambda (e)
    (tc:type-parse (tc:unparse-input-type-exp e))))

(define replace-variables
  (lambda (var-names var-types type-expr)
    (if (null? var-names)
	type-expr
	(replace-variables
	 (cdr var-names)
	 (cdr var-types)
	 (tc:type-expr-replace (tc:basic-type-expr (car var-names))
			       (car var-types)
			       type-expr)))))	      

(deftype tc:mult-def-down-type
  (-> ((list-of (pair-of tc:input-type-exp tc:input-type-exp)) tc:input-type-exp)
      tc:input-type-exp))

(define tc:mult-def-down-type
  (lambda (defreps typ)
    (if (null? defreps)
	typ
	(let* ((internal-defreps
		(map

		 (lambda (p)
		   (cases tc:scheme-abs-conc-pair p
			  (tc:abs-conc-pair-simple
			   (position abstract-type concrete-type)
			   (cons (tc:infer-input-type-exp abstract-type)
				 (tc:infer-input-type-exp concrete-type)))
			  (tc:abs-conc-pair-polymorphic
			   (position type-vars abstract-type concrete-type)
			   (let* ((var-tes (tc:listof-new-variable-type-expr (length type-vars)))
				  (abs-t (replace-variables type-vars var-tes
							    (tc:infer-input-type-exp abstract-type)))
				  (conc-t (replace-variables type-vars var-tes
							     (tc:infer-input-type-exp concrete-type))))
			     (cons abs-t conc-t)))))
		     defreps))
	       (internal-typ (tc:infer-input-type-exp typ))
	       (down-internal-typ
		(tc:internal-mult-def-down-type internal-defreps
						internal-typ)))
	  ;;(displayln "tc:internal-mult-def-down-type returned "
	  ;;           (tc:type-unparse down-internal-typ))
	  (tc:translated-input-type-exp
	   (tc:input-type-exp->position typ)
	   down-internal-typ)))))

(deftype tc:internal-mult-def-down-type
  (-> ((list-of (pair-of tc:type-expr tc:type-expr)) tc:type-expr)
      tc:type-expr))
(define tc:internal-mult-def-down-type
  (lambda (defreps typ)
;;    (displayln "DEFREPS:")
;;    (map displayln defreps)
    
    (if (null? defreps)
	typ
	(tc:internal-mult-def-down-type (cdr defreps)
					(tc:down-type (car defreps) typ)))))

(deftype tc:down-type 
  (-> ((pair-of tc:type-expr tc:type-expr) tc:type-expr)
      tc:type-expr))
(define tc:down-type
  (lambda (abst-rep-pair declared-type)
    ;; ENSURES: result is a tc:type-expr like declared-type but in which
    ;;          each sub-expression that unifies with the car of abst-rep-pair
    ;;          has been substituted by the cdr of abst-rep-pair under the
    ;;          unifying substitution.
    ;;(displayln "tc:down-type: abst-rep-pair = " abst-rep-pair)

;;    (displayln "declared-type: " declared-type)
;;    (displayln "abs part: " (car abst-rep-pair))
;;    (displayln "concpart: " (cdr abst-rep-pair))
    
    (let* ((result (tc:deep-down-type (tc:type-expr-uniquify declared-type)
				      abst-rep-pair
				      (tc:type-expr-null-subst)))
	   (texpr (car result))
	   (env (cdr result)))
      ;; (displayln "texpr = " texpr)
      ;; (displayln "env = " env)
      ((tc:app env) texpr))))

;;[[[FIXME: fix all the types and variant stuff?]]]
(deftype tc:deep-down-type
  (-> (tc:type-expr (pair-of tc:type-expr tc:type-expr)
		    (tc:env-of uniq-sym tc:type-expr))
      (pair-of tc:type-expr (tc:env-of uniq-sym tc:type-expr))))
(define tc:deep-down-type
  (lambda (up-type abst-rep-pair env)
    (let* ((uniq-rep (tc:type-expr-uniquify abst-rep-pair))
	   (abst-type (car uniq-rep))
	   (rep-type (cdr uniq-rep)))
;;       (displayln "abst-type: " abst-type)
;;       (displayln "rep-type: " rep-type)
      (letrec
	  ((down
	    (lambda (up-type env)
	      (cond
	       ((tc:variable-type-expr? up-type) (cons up-type env))
	       ((tc:basic-type-expr? up-type)
		(cons (if (and (tc:basic-type-expr? abst-type)
			       (tc:basic-type-expr-eq? up-type abst-type))
			  rep-type
			  up-type)
		      env))
	       ((tc:function-type-expr? up-type)
		(let* ((res-args
			(down-type-list
			 (tc:function-type-expr->arg-types up-type) env))
		       (res-result
			(down
			 (tc:function-type-expr->result-type up-type)
			 (cdr res-args))))
		  (unify-and-subst
		   (tc:make-function-type-expr (car res-args)
					       (car res-result))
		   (cdr res-result))))
	       ((tc:intersection-type-expr? up-type)
		(let ((res-conj (down-type-list
				 (tc:intersection-type-expr->conjoined-types
				  up-type) env)))
		  (unify-and-subst
		   (tc:make-intersection-type-expr (car res-conj))
		   (cdr res-conj))))
	       ((tc:applied-type-expr? up-type)
		(let* ((res-rator
			(down
			 (tc:applied-type-expr->operator-type up-type)
			 env))
		       (res-rands
			(has-type
			 (list-of (list-of type-expr))
			 (down-type-list
			  (tc:applied-type-expr->operand-types up-type)
			  (cdr res-rator)))))
		  (unify-and-subst
		   (tc:make-applied-type-expr
		    (car res-rator)
		    (car res-rands))
		   (cdr res-rands))))
	       ((tc:type-predicate-for-type-expr? up-type)
		(let ((res (down (tc:type-predicate-for-type-expr->type up-type)
				 env)))
		  (unify-and-subst
		   (tc:make-type-predicate-for-type-expr (car res))
		   (cdr res))))
	       (else (error
		      "tc:deep-down-type: invalid type expr: " up-type)))))
	   (down-type-list
	    (has-type
	     (-> ((list-of tc:type-expr) (tc:env-of uniq-sym tc:type-expr))
		 (pair-of (list-of tc:type-expr)
			  (tc:env-of uniq-sym tc:type-expr)))
	     (lambda (up-types new-env)
	       ;; Each tc:type-expr in the list is unified in an
	       ;; environment formed by the unification of the preceding
	       ;; tc:type-exprs in the list.
	       (if (null? up-types)
		   (cons '() new-env)
		   (let* ((res-first (down (car up-types) new-env))
			  (res-rest (down-type-list (cdr up-types)
						    (cdr res-first))))
		     (cons (cons (car res-first) (car res-rest))
			   (cdr res-rest)))))))
	   (unify-and-subst
	    (has-type
	     (-> (tc:type-expr (tc:env-of uniq-sym tc:type-expr))
		 (pair-of tc:type-expr
			  (list-of tc:subst)))
	     (lambda (ut ne)
	       (let ((mgu-result ((tc:mgu (tc:type-expr-as-unifiable)) ut abst-type)))
		 (if (null? mgu-result)
		     ;; If unification failed then return type-expr
		     ;; and environment without changes
		     (cons ut ne)
		     ;; If unification succeeded then replace the
		     ;; tc:type-expr by the representation of the abstract
		     ;; type with bindings substituted
		     (cons
		      ((tc:app (car mgu-result)) rep-type)
		      (car mgu-result))))))))
	(down up-type env)))))

) ;; End module
